import asyncio
import copy
import re
from collections import defaultdict
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Type

import grpc.aio

from jina._docarray import DocumentArray, docarray_v2
from jina.constants import __default_endpoint__
from jina.excepts import InternalNetworkError
from jina.logging.logger import JinaLogger
from jina.serve.networking import GrpcConnectionPool
from jina.serve.runtimes.helper import _parse_specific_params
from jina.serve.runtimes.worker.request_handling import WorkerRequestHandler
from jina.types.request.data import DataRequest, SingleDocumentRequest

if docarray_v2:
    from docarray import DocList
    from docarray.documents.legacy import LegacyDocument

    from jina.serve.runtimes.helper import _create_pydantic_model_from_schema

    legacy_doc_schema = LegacyDocument.schema()


class TopologyGraph:
    """
    :class TopologyGraph is a class that describes a computational graph of nodes, where each node represents
        a Deployment that needs to be sent requests in the order respecting the path traversal.

    :param graph_description: A dictionary describing the topology of the Deployments. 2 special nodes are expected, the name `start-gateway` and `end-gateway` to
        determine the nodes that receive the very first request and the ones whose response needs to be sent back to the client. All the nodes with no outgoing nodes
        will be considered to be floating, and they will be "flagged" so that the user can ignore their tasks and not await them.

    :param conditions: A dictionary describing which Executors have special conditions to be fullfilled by the `Documents` to be sent to them.
    :param reduce: Reduce requests arriving from multiple needed predecessors, True by default
    """

    class _ReqReplyNode:
        def __init__(
                self,
                name: str,
                number_of_parts: int = 1,
                floating: bool = False,
                filter_condition: dict = None,
                metadata: Optional[Dict] = None,
                reduce: bool = True,
                timeout_send: Optional[float] = None,
                retries: Optional[int] = -1,
                logger: Optional[JinaLogger] = None,
        ):
            self.name = name
            self.outgoing_nodes = []
            self.number_of_parts = number_of_parts
            self.floating = floating
            self.parts_to_send = []
            self.start_time = None
            self.end_time = None
            self.status = None
            self._filter_condition = filter_condition
            self._metadata = metadata
            self._reduce = reduce
            self._timeout_send = timeout_send
            self._retries = retries
            self.result_in_params_returned = None
            self.logger = logger or JinaLogger(self.__class__.__name__)
            self.endpoints = None
            self._pydantic_models_by_endpoint = None
            self._endpoints_proto = None

        @property
        def leaf(self):
            return len(self.outgoing_nodes) == 0

        def _validate_against_outgoing_nodes(self):
            def _check_schema_equality(schema_1, schema_2):
                from collections import OrderedDict
                # Naive check of compatibility
                schema_1_properties = OrderedDict(sorted(schema_1.get('properties', {}).items()))
                schema_2_properties = OrderedDict(sorted(schema_2.get('properties', {}).items()))
                if len(schema_1_properties) != len(schema_2_properties):
                    return False

                for property_1, property_2 in zip(schema_1_properties.keys(), schema_2_properties.keys()):
                    if property_1 != property_2:
                        return False

                    if schema_1_properties[property_1].get('type', None) != schema_2_properties[property_2].get('type', None):
                        return False

                    # TODO: Add more complex check for nested definitions

                return True

            for node in self.outgoing_nodes:
                # here validate for each endpoint that output of self matches input of node
                if node._pydantic_models_by_endpoint is not None:  # gateway end

                    for endp in self._pydantic_models_by_endpoint.keys():
                        outgoing_endp = endp
                        incoming_endp = (
                            endp
                            if endp in node._pydantic_models_by_endpoint
                            else __default_endpoint__
                        )

                        if incoming_endp in node._pydantic_models_by_endpoint:

                            if endp in node._pydantic_models_by_endpoint:
                                if not _check_schema_equality(self._pydantic_models_by_endpoint[outgoing_endp][
                                                                  'output'
                                                              ].schema(),
                                                              node._pydantic_models_by_endpoint[incoming_endp][
                                                                  'input'
                                                              ].schema()):
                                    raise Exception(
                                        f'The output schema of {self.name} at {outgoing_endp} endpoint is incompatible with the input schema of {node.name} at {incoming_endp} endpoint'
                                    )
                        else:
                            if (
                                    outgoing_endp != __default_endpoint__
                            ):  # It could happen that there is an Encoder with default followed by an indexer with [index, search]
                                raise Exception(
                                    f'{node.name} does not expose {incoming_endp} which makes it impossible to be chained with {self.name} on {outgoing_endp}'
                                )
                            else:
                                self.logger.warning(
                                    f'{node.name} does not expose {incoming_endp} which could lead to incompatibility when calling non-explicitly bound endpoints'
                                )
                return node._validate_against_outgoing_nodes()
            return True

        def _update_requests_with_filter_condition(self, need_copy):
            for i in range(len(self.parts_to_send)):
                doc_array_cls = self.parts_to_send[i].document_array_cls
                req = (
                    self.parts_to_send[i]
                    if not need_copy
                    else copy.deepcopy(self.parts_to_send[i])
                )
                # somehow the document_array_cls is not copied
                req.document_array_cls = doc_array_cls
                if not docarray_v2:
                    filtered_docs = req.docs.find(self._filter_condition)
                else:
                    from docarray.utils.filter import filter_docs

                    filtered_docs = filter_docs(req.docs, self._filter_condition)
                req.data.docs = filtered_docs
                self.parts_to_send[i] = req

        def _update_request_by_params(
                self, deployment_name: str, request_input_parameters: Dict
        ):
            specific_parameters = _parse_specific_params(
                request_input_parameters, deployment_name
            )
            for i in range(len(self.parts_to_send)):
                self.parts_to_send[i].parameters = specific_parameters

        def _handle_internalnetworkerror(self, err):
            err_code = err.code()
            if err_code == grpc.StatusCode.UNAVAILABLE:
                err._details = (
                        err.details()
                        + f' |Gateway: Communication error with deployment {self.name} at address(es) {err.dest_addr}. '
                          f'Head or worker(s) may be down.'
                )
                raise err
            elif err_code == grpc.StatusCode.DEADLINE_EXCEEDED:
                err._details = (
                        err.details()
                        + f'|Gateway: Connection with deployment {self.name} at address(es) {err.dest_addr} could be established, but timed out.'
                          f' You can increase the allowed time by setting `timeout_send` in your Flow YAML `with` block or Flow `__init__()` method.'
                )
                raise err
            elif err_code == grpc.StatusCode.NOT_FOUND:
                err._details = (
                        err.details()
                        + f'\n|Gateway: Connection error with deployment `{self.name}` at address(es) {err.dest_addr}.'
                          f' Connection with {err.dest_addr} succeeded, but `{self.name}` was not found.'
                          f' Possibly `{self.name}` is behind an API gateway but not reachable.'
                )
                raise err
            else:
                raise

        def get_endpoints(
                self,
                connection_pool: GrpcConnectionPool,
                models_schema_list: List,
                models_list: List,
        ) -> asyncio.Task:
            # models_schema_list and models_list is given to each node. And each one fills its models
            from google.protobuf import json_format

            async def task():
                if self._endpoints_proto is None:
                    self.logger.debug(f'Getting Endpoints data from {self.name}')
                    endpoints_proto = await connection_pool.send_discover_endpoint(
                        self.name, retries=self._retries
                    )
                    # TODO: Try more often should not be able to start, synchronization issue
                    if endpoints_proto is not None:
                        endp, _ = endpoints_proto
                        self.endpoints = endp.endpoints
                        if docarray_v2:
                            from docarray.documents.legacy import LegacyDocument

                            schemas = json_format.MessageToDict(endp.schemas)
                            self._pydantic_models_by_endpoint = {}
                            models_created_by_name = {}
                            for endpoint, inner_dict in schemas.items():
                                input_model_name = inner_dict['input']['name']
                                input_model_schema = inner_dict['input']['model']
                                if input_model_schema in models_schema_list:
                                    input_model = models_list[
                                        models_schema_list.index(input_model_schema)
                                    ]
                                    models_created_by_name[
                                        input_model_name
                                    ] = input_model
                                else:
                                    if input_model_name not in models_created_by_name:
                                        if input_model_schema == legacy_doc_schema:
                                            input_model = LegacyDocument
                                        else:
                                            input_model = (
                                                _create_pydantic_model_from_schema(
                                                    input_model_schema,
                                                    input_model_name,
                                                    models_created_by_name,
                                                )
                                            )
                                        models_created_by_name[
                                            input_model_name
                                        ] = input_model
                                    input_model = models_created_by_name[
                                        input_model_name
                                    ]
                                    models_schema_list.append(input_model_schema)
                                    models_list.append(input_model)

                                output_model_name = inner_dict['output']['name']
                                output_model_schema = inner_dict['output']['model']
                                if output_model_schema in models_schema_list:
                                    output_model = models_list[
                                        models_schema_list.index(output_model_schema)
                                    ]
                                    models_created_by_name[
                                        output_model_name
                                    ] = output_model
                                else:
                                    if output_model_name not in models_created_by_name:
                                        if output_model_name == legacy_doc_schema:
                                            output_model = LegacyDocument
                                        else:
                                            output_model = (
                                                _create_pydantic_model_from_schema(
                                                    output_model_schema,
                                                    output_model_name,
                                                    models_created_by_name,
                                                )
                                            )
                                        models_created_by_name[
                                            output_model_name
                                        ] = output_model
                                    output_model = models_created_by_name[
                                        output_model_name
                                    ]
                                    models_schema_list.append(output_model)
                                    models_list.append(output_model)

                                parameters_model_name = inner_dict['parameters']['name']
                                parameters_model_schema = inner_dict['parameters'][
                                    'model'
                                ]
                                if parameters_model_schema is not None:
                                    if parameters_model_schema in models_schema_list:
                                        parameters_model = models_list[
                                            models_schema_list.index(
                                                parameters_model_schema
                                            )
                                        ]
                                        models_created_by_name[
                                            parameters_model_name
                                        ] = parameters_model
                                    else:
                                        if (
                                                parameters_model_name
                                                not in models_created_by_name
                                        ):
                                            from pydantic import BaseModel

                                            parameters_model = (
                                                _create_pydantic_model_from_schema(
                                                    parameters_model_schema,
                                                    parameters_model_name,
                                                    models_created_by_name,
                                                    base_class=BaseModel,
                                                )
                                            )
                                            models_created_by_name[
                                                parameters_model_name
                                            ] = parameters_model
                                    parameters_model = models_created_by_name[
                                        parameters_model_name
                                    ]
                                    models_schema_list.append(parameters_model_schema)
                                    models_list.append(parameters_model)
                                else:
                                    parameters_model = None

                                self._pydantic_models_by_endpoint[endpoint] = {
                                    'input': input_model,
                                    'output': output_model,
                                    'is_generator': inner_dict['is_generator'],
                                    'is_singleton_doc': inner_dict['is_singleton_doc'],
                                    'parameters': parameters_model,
                                }
                        self._endpoints_proto = endpoints_proto
                    else:
                        raise Exception('Failed to get endpoints')
                return self._endpoints_proto

            return asyncio.create_task(task())

        async def stream_single_doc(
                self,
                request: SingleDocumentRequest,
                connection_pool: GrpcConnectionPool,
                endpoint: Optional[str],
                return_type: Type[DocumentArray] = DocumentArray,
        ):
            if docarray_v2:
                if self.endpoints and endpoint in self.endpoints:
                    request.document_cls = self._pydantic_models_by_endpoint[endpoint][
                        'input'
                    ]

            async for resp, _ in connection_pool.send_single_document_request(
                    request=request,
                    deployment=self.name,
                    metadata=self._metadata,
                    head=True,
                    endpoint=endpoint,
                    timeout=self._timeout_send,
                    retries=self._retries,
            ):
                if issubclass(type(resp), BaseException):
                    raise resp
                else:
                    if docarray_v2:
                        if self.endpoints and endpoint in self.endpoints:
                            from docarray.base_doc import AnyDoc

                            # if return_type is not specified or if it is a default type, cast using retrieved
                            # schemas
                            if (
                                    not return_type
                                    or not return_type.doc_type
                                    or return_type.doc_type is AnyDoc
                            ):
                                resp.document_cls = self._pydantic_models_by_endpoint[
                                    endpoint
                                ]['output']
                            else:
                                resp.document_array_cls = return_type
                    yield resp

        async def _wait_previous_and_send(
                self,
                request: Optional[DataRequest],
                previous_task: Optional[asyncio.Task],
                connection_pool: GrpcConnectionPool,
                endpoint: Optional[str],
                target_executor_pattern: Optional[str] = None,
                request_input_parameters: Dict = {},
                copy_request_at_send: bool = False,
                init_task: Optional[asyncio.Task] = None,
                return_type: Type[DocumentArray] = None,
        ):
            # Check my condition and send request with the condition
            metadata = {}
            if init_task is not None:
                await init_task
            if previous_task is not None:
                result = await previous_task
                request, metadata = result[0], result[1]
            if metadata and 'is-error' in metadata:
                return request, metadata
            elif request is not None:

                request.parameters = _parse_specific_params(
                    request.parameters, self.name
                )
                req_to_send = (
                    copy.deepcopy(request) if copy_request_at_send else request
                )
                if docarray_v2:
                    if self.endpoints and endpoint in self.endpoints:
                        req_to_send.document_array_cls = DocList[
                            self._pydantic_models_by_endpoint[endpoint]['input']
                        ]

                self.parts_to_send.append(req_to_send)
                # this is a specific needs
                if len(self.parts_to_send) == self.number_of_parts:
                    self.start_time = datetime.utcnow()
                    self._update_request_by_params(self.name, request_input_parameters)
                    if self._filter_condition is not None:
                        self._update_requests_with_filter_condition(
                            need_copy=not copy_request_at_send
                        )

                    # avoid sending to executor which does not bind to this endpoint
                    if endpoint is not None and self.endpoints is not None:
                        if (
                                endpoint not in self.endpoints
                                and __default_endpoint__ not in self.endpoints
                        ):
                            return request, metadata

                    if self._reduce and len(self.parts_to_send) > 1:
                        self.parts_to_send = [
                            WorkerRequestHandler.reduce_requests(self.parts_to_send)
                        ]

                    if target_executor_pattern is not None and not re.match(
                            target_executor_pattern, self.name
                    ):
                        return request, metadata
                    # otherwise, send to executor and get response
                    try:
                        result = await connection_pool.send_requests_once(
                            requests=self.parts_to_send,
                            deployment=self.name,
                            metadata=self._metadata,
                            head=True,
                            endpoint=endpoint,
                            timeout=self._timeout_send,
                            retries=self._retries,
                        )
                        if issubclass(type(result), BaseException):
                            raise result
                        else:
                            resp, metadata = result

                        if docarray_v2:
                            if self.endpoints and (
                                    endpoint in self.endpoints
                                    or __default_endpoint__ in self.endpoints
                            ):
                                from docarray.base_doc import AnyDoc

                                # if return_type is not specified or if it is a default type, cast using retrieved
                                # schemas
                                if (
                                        not return_type
                                        or not return_type.doc_type
                                        or return_type.doc_type is AnyDoc
                                ):
                                    pydantic_models = (
                                            self._pydantic_models_by_endpoint.get(endpoint)
                                            or self._pydantic_models_by_endpoint.get(
                                        __default_endpoint__
                                    )
                                    )
                                    resp.document_array_cls = DocList[
                                        pydantic_models['output']
                                    ]
                                else:
                                    resp.document_array_cls = return_type

                        if WorkerRequestHandler._KEY_RESULT in resp.parameters:
                            # Accumulate results from each Node and then add them to the original
                            self.result_in_params_returned = resp.parameters[
                                WorkerRequestHandler._KEY_RESULT
                            ]
                        request.parameters = request_input_parameters
                        resp.parameters = request_input_parameters
                        self.parts_to_send.clear()
                    except InternalNetworkError as err:
                        self._handle_internalnetworkerror(err)
                    except Exception as err:
                        self.logger.error(
                            f'Exception sending requests to {self.name}: {err}'
                        )
                        raise err

                    self.end_time = datetime.utcnow()
                    if metadata and 'is-error' in metadata:
                        self.status = resp.header.status
                    return resp, metadata

            return None, {}

        def _get_input_output_model_for_endpoint(
                self,
                previous_input,
                previous_output,
                previous_is_generator,
                previous_is_singleton_doc,
                previous_parameters,
                endpoint,
        ):
            if self._pydantic_models_by_endpoint is not None:

                if endpoint in self.endpoints:
                    # update output
                    # the only important for the gateway to know if is singleton is the first (so the last/ ignore previous is_singleton)
                    # the only important for the gateway to know if is generator is the last (so the first/previous precedence)
                    new_input = previous_input
                    if previous_input is None:
                        new_input = self._pydantic_models_by_endpoint[endpoint]['input']

                    is_generator = previous_is_generator
                    if previous_is_generator is None:
                        is_generator = self._pydantic_models_by_endpoint[endpoint][
                            'is_generator'
                        ]

                    if (
                            previous_output
                            and previous_output.schema()
                            == self._pydantic_models_by_endpoint[endpoint][
                        "output"
                    ].schema()
                    ):
                        # this is needed to not mix model IDs, otherwise FastAPI gets crazy
                        return {
                            'input': new_input,
                            'output': previous_output,
                            'is_generator': is_generator,
                            'is_singleton_doc': self._pydantic_models_by_endpoint[
                                endpoint
                            ]['is_singleton_doc'],
                            'parameters': self._pydantic_models_by_endpoint[endpoint][
                                'parameters'
                            ],
                        }
                    else:
                        return {
                            'input': new_input,
                            'output': self._pydantic_models_by_endpoint[endpoint][
                                'output'
                            ],
                            'is_generator': is_generator,
                            'is_singleton_doc': self._pydantic_models_by_endpoint[
                                endpoint
                            ]['is_singleton_doc'],
                            'parameters': self._pydantic_models_by_endpoint[endpoint][
                                'parameters'
                            ],
                        }
                else:
                    return {
                        'input': previous_input,
                        'output': previous_output,
                        'is_generator': previous_is_generator,
                        'is_singleton_doc': False,
                        'parameters': previous_parameters,
                    }
            return None

        def _get_leaf_input_output_model(
                self,
                previous_input,
                previous_output,
                previous_is_generator,
                previous_is_singleton_doc,
                previous_parameters,
                endpoint: Optional[str] = None,
        ):
            new_map = self._get_input_output_model_for_endpoint(
                previous_input,
                previous_output,
                previous_is_generator,
                previous_is_singleton_doc,
                previous_parameters,
                endpoint,
            )
            if self.leaf:  # I am like a leaf
                return list(
                    [new_map] if new_map is not None else []
                )  # I am the last in the chain
            list_of_outputs = []
            for outgoing_node in self.outgoing_nodes:
                list_of_maps = outgoing_node._get_leaf_input_output_model(
                    previous_input=new_map['input'] if new_map is not None else None,
                    previous_output=new_map['output'] if new_map is not None else None,
                    previous_is_generator=new_map['is_generator']
                    if new_map is not None
                    else None,
                    previous_is_singleton_doc=new_map['is_singleton_doc']
                    if new_map is not None
                    else None,
                    previous_parameters=new_map['parameters']
                    if new_map is not None
                    else None,
                    endpoint=endpoint,
                )
                # We are interested in the last one, that will be the task that awaits all the previous
                list_of_outputs.extend(list_of_maps)

            return list_of_outputs

        def get_leaf_req_response_tasks(
                self,
                connection_pool: GrpcConnectionPool,
                request_to_send: Optional[DataRequest],
                previous_task: Optional[asyncio.Task],
                endpoint: Optional[str] = None,
                target_executor_pattern: Optional[str] = None,
                request_input_parameters: Dict = {},
                request_input_has_specific_params: bool = False,
                copy_request_at_send: bool = False,
                init_task: Optional[asyncio.Task] = None,
                return_type: Type[DocumentArray] = DocumentArray,
        ) -> List[Tuple[bool, asyncio.Task]]:
            """
            Gets all the tasks corresponding from all the subgraphs born from this node

            :param connection_pool: The connection_pool need to actually send the requests
            :param request_to_send: Optional request to be sent when the node is an origin of a graph
            :param previous_task: Optional task coming from the predecessor of the Node
            :param endpoint: Optional string defining the endpoint of this request
            :param target_executor_pattern: Optional regex pattern for the target executor to decide whether or not the Executor should receive the request
            :param request_input_parameters: The parameters coming from the Request as they arrive to the gateway
            :param request_input_has_specific_params: Parameter added for optimization. If this is False, there is no need to copy at all the request
            :param copy_request_at_send: Copy the request before actually calling the `ConnectionPool` sending
            :param init_task: Initial task to be called before sending any request

            .. note:
                deployment1 -> outgoing_nodes: deployment2
                deployment2 -> outgoing_nodes: deployment4
                deployment3 -> outgoing_nodes: deployment4
                deployment4 -> outgoing_nodes: deployment6
                deployment5 -> outgoing_nodes: deployment6
                deployment6 -> outgoing_nodes: []

                |-> deployment1 -> deployment2 -->
                |                   | -> deployment4 --->
                |-> deployment3 ---------->             | -> deployment6
                |-> deployment5 ------------------------>

                Let's imagine a graph from this. Node corresponding to Deployment6 will receive 2 calls from deployment4 and deployment5.
                The task returned by `deployment6` will backpropagated to the caller of deployment1.get_leaf_tasks, deployment3.get_leaf_tasks and deployment5.get_leaf_tasks.

                When the caller of these methods await them, they will fire the logic of sending requests and responses from and to every deployment

            :param return_type: the DocumentArray type to be returned. By default, it is `DocumentArray`.
            :return: Return a list of tuples, where tasks corresponding to the leafs of all the subgraphs born from this node are in each tuple.
                These tasks will be based on awaiting for the task from previous_node and sending a request to the corresponding node. The other member of the pair
                is a flag indicating if the task is to be awaited by the gateway or not.
            """
            wait_previous_and_send_task = asyncio.create_task(
                self._wait_previous_and_send(
                    request=request_to_send,
                    previous_task=previous_task,
                    connection_pool=connection_pool,
                    endpoint=endpoint,
                    target_executor_pattern=target_executor_pattern,
                    request_input_parameters=request_input_parameters,
                    copy_request_at_send=copy_request_at_send,
                    init_task=init_task,
                    return_type=return_type,
                )
            )
            if self.leaf:  # I am like a leaf
                return [
                    (not self.floating, wait_previous_and_send_task)
                ]  # I am the last in the chain
            hanging_tasks_tuples = []
            num_outgoing_nodes = len(self.outgoing_nodes)
            for outgoing_node in self.outgoing_nodes:
                t = outgoing_node.get_leaf_req_response_tasks(
                    connection_pool=connection_pool,
                    request_to_send=None,
                    previous_task=wait_previous_and_send_task,
                    endpoint=endpoint,
                    target_executor_pattern=target_executor_pattern,
                    request_input_parameters=request_input_parameters,
                    request_input_has_specific_params=request_input_has_specific_params,
                    copy_request_at_send=num_outgoing_nodes > 1
                                         and request_input_has_specific_params,
                    return_type=return_type,
                )
                # We are interested in the last one, that will be the task that awaits all the previous
                hanging_tasks_tuples.extend(t)

            return hanging_tasks_tuples

        def add_route(self, request: 'DataRequest'):
            """
             Add routes to the DataRequest based on the state of request processing

             :param request: the request to add the routes to
            :return: modified request with added routes
            """

            def _find_route(request):
                for r in request.routes:
                    if r.executor == self.name:
                        return r
                return None

            r = _find_route(request)
            if r is None and self.start_time:
                r = request.routes.add()
                r.executor = self.name
                r.start_time.FromDatetime(self.start_time)
                if self.end_time:
                    r.end_time.FromDatetime(self.end_time)
                if self.status:
                    r.status.CopyFrom(self.status)
            for outgoing_node in self.outgoing_nodes:
                request = outgoing_node.add_route(request=request)
            return request

    class _EndGatewayNode(_ReqReplyNode):

        """
        Dummy node to be added before the gateway. This is to solve a problem we had when implementing `floating Executors`.
        If we do not add this at the end, this structure does not work:

            GATEWAY -> EXEC1 -> FLOATING
                    -> GATEWAY
        """

        def get_endpoints(self, *args, **kwargs) -> asyncio.Task:
            async def task_wrapper():
                from jina.serve.networking import default_endpoints_proto

                return default_endpoints_proto, None

            return asyncio.create_task(task_wrapper())

        def get_leaf_req_response_tasks(
                self, previous_task: Optional[asyncio.Task], *args, **kwargs
        ) -> List[Tuple[bool, asyncio.Task]]:
            return [(True, previous_task)]

        def _get_leaf_input_output_model(
                self,
                previous_input,
                previous_output,
                previous_is_generator,
                previous_is_singleton_doc,
                previous_parameters,
                endpoint: Optional[str] = None,
        ):
            return [
                {
                    'input': previous_input,
                    'output': previous_output,
                    'is_generator': previous_is_generator,
                    'is_singleton_doc': previous_is_singleton_doc,
                    'parameters': previous_parameters,
                }
            ]

    def __init__(
            self,
            graph_representation: Dict,
            graph_conditions: Dict = {},
            deployments_metadata: Dict = {},
            deployments_no_reduce: List[str] = [],
            timeout_send: Optional[float] = 1.0,
            retries: Optional[int] = -1,
            logger: Optional[JinaLogger] = None,
            *args,
            **kwargs,
    ):
        self.logger = logger or JinaLogger(self.__class__.__name__)
        num_parts_per_node = defaultdict(int)
        if 'start-gateway' in graph_representation:
            origin_node_names = graph_representation['start-gateway']
        else:
            origin_node_names = set()
        floating_deployment_set = set()
        node_set = set()
        for node_name, outgoing_node_names in graph_representation.items():
            if node_name not in {'start-gateway', 'end-gateway'}:
                node_set.add(node_name)
            if len(outgoing_node_names) == 0:
                floating_deployment_set.add(node_name)
            for out_node_name in outgoing_node_names:
                if out_node_name not in {'start-gateway', 'end-gateway'}:
                    node_set.add(out_node_name)
                    num_parts_per_node[out_node_name] += 1

        nodes = {}
        for node_name in node_set:
            condition = graph_conditions.get(node_name, None)
            metadata = deployments_metadata.get(node_name, None)
            nodes[node_name] = self._ReqReplyNode(
                name=node_name,
                number_of_parts=num_parts_per_node[node_name]
                if num_parts_per_node[node_name] > 0
                else 1,
                floating=node_name in floating_deployment_set,
                filter_condition=condition,
                metadata=metadata,
                reduce=node_name not in deployments_no_reduce,
                timeout_send=timeout_send,
                retries=retries,
                logger=self.logger,
            )

        for node_name, outgoing_node_names in graph_representation.items():
            if node_name not in ['start-gateway', 'end-gateway']:
                for out_node_name in outgoing_node_names:
                    if out_node_name not in ['start-gateway', 'end-gateway']:
                        nodes[node_name].outgoing_nodes.append(nodes[out_node_name])
                    if out_node_name == 'end-gateway':
                        nodes[node_name].outgoing_nodes.append(
                            self._EndGatewayNode(name='__end_gateway__', floating=False)
                        )

        self._origin_nodes = [nodes[node_name] for node_name in origin_node_names]
        self.has_filter_conditions = bool(graph_conditions)
        self._all_endpoints = None

    async def _get_all_endpoints(
            self, connection_pool, retry_forever=False, is_cancel=None
    ):
        def _condition():
            if is_cancel is not None:
                is_cancelled = is_cancel.is_set()
                if is_cancelled:
                    self.logger.debug(f'cancel get all endpoints')
                return not is_cancelled
            else:
                return True

        if not self._all_endpoints:
            while _condition():
                try:
                    models_schemas_list = []
                    models_list = []
                    tasks_to_get_endpoints = [
                        node.get_endpoints(
                            connection_pool, models_schemas_list, models_list
                        )
                        for node in self.all_nodes
                    ]
                    await asyncio.gather(*tasks_to_get_endpoints)
                    endpoints = set()
                    for node in self.all_nodes:
                        if node.endpoints is not None:
                            endpoints.update(list(node.endpoints))
                    self._all_endpoints = endpoints
                    break
                except Exception as exc:
                    if not retry_forever:
                        raise exc
                    self.logger.warning(
                        f'Getting endpoints failed: {exc}. Waiting for another trial'
                    )
                    await asyncio.sleep(1)

        return self._all_endpoints

    def add_routes(self, request: 'DataRequest'):
        """
        Add routes to the DataRequest based on the state of request processing

        :param request: the request to add the routes to
        :return: modified request with added routes
        """
        for node in self._origin_nodes:
            request = node.add_route(request=request)
        return request

    @property
    def origin_nodes(self):
        """
        The list of origin nodes, the one that depend only on the gateway, so all the subgraphs will be born from them and they will
        send to their deployments the request as received by the client.

        :return: A list of nodes
        """
        return self._origin_nodes

    @property
    def all_nodes(self):
        """
        The set of all the nodes inside this Graph

        :return: A list of nodes
        """

        def _get_all_nodes(node, accum, accum_names):
            if node.name not in accum_names:
                accum.append(node)
                accum_names.append(node.name)
            for n in node.outgoing_nodes:
                _get_all_nodes(n, accum, accum_names)
            return accum, accum_names

        nodes = []
        node_names = []
        for origin_node in self.origin_nodes:
            subtree_nodes, subtree_node_names = _get_all_nodes(origin_node, [], [])
            for st_node, st_node_name in zip(subtree_nodes, subtree_node_names):
                if st_node_name not in node_names:
                    nodes.append(st_node)
                    node_names.append(st_node_name)
        return nodes

    def collect_all_results(self):
        """Collect all the results from every node into a single dictionary so that gateway can collect them

        :return: A dictionary of the results
        """
        res = {}
        for node in self.all_nodes:
            if node.result_in_params_returned:
                res.update(node.result_in_params_returned)
        return res

    def _validate_flow_docarray_compatibility(self):
        """
        Validates flow docarray validity in terms of input-output schemas of Executors
        """
        for node in self.origin_nodes:
            node._validate_against_outgoing_nodes()



from jina.parsers.helper import KVAppendAction


def mixin_grpc_channel_options_parser(arg_group):
    """Mixing in arguments required by any class that extends :class:`AsynNewLoopRuntime` into the given parser.
    :param arg_group: the parser instance to which we add arguments
    """

    arg_group.add_argument(
        '--grpc-channel-options',
        action=KVAppendAction,
        metavar='KEY: VALUE',
        nargs='*',
        help="Dictionary of kwargs arguments that will be passed to the grpc channel as options when creating a channel, example : {'grpc.max_send_message_length': -1}. When max_attempts > 1, the 'grpc.service_config' option will not be applicable.",
        default=None,
    )



try:
    from docarray import BaseDoc as Document
    from docarray import DocList as DocumentArray

    docarray_v2 = True

except ImportError:
    from docarray import Document, DocumentArray

    docarray_v2 = False



import asyncio
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Set, Tuple, Union, AsyncGenerator

import grpc
from grpc.aio import AioRpcError

from jina.constants import __default_endpoint__
from jina.enums import PollingType
from jina.excepts import InternalNetworkError
from jina.importer import ImportExtensions
from jina.logging.logger import JinaLogger
from jina.proto import jina_pb2
from jina.serve.helper import format_grpc_error
from jina.serve.networking.connection_pool_map import _ConnectionPoolMap
from jina.serve.networking.connection_stub import create_async_channel_stub
from jina.serve.networking.instrumentation import (
    _NetworkingHistograms,
    _NetworkingMetrics,
)
from jina.serve.networking.replica_list import _ReplicaList
from jina.serve.networking.utils import DEFAULT_MINIMUM_RETRIES
from jina.types.request import Request
from jina.types.request.data import SingleDocumentRequest

if TYPE_CHECKING:  # pragma: no cover
    from grpc.aio._interceptor import ClientInterceptor
    from opentelemetry.instrumentation.grpc._client import (
        OpenTelemetryClientInterceptor,
    )
    from opentelemetry.metrics import Meter
    from prometheus_client import CollectorRegistry, Summary

default_endpoints_proto = jina_pb2.EndpointsProto()
default_endpoints_proto.endpoints.extend([__default_endpoint__])


class GrpcConnectionPool:
    """
    Manages a list of grpc connections.

    :param logger: the logger to use
    :param compression: The compression algorithm to be used by this GRPCConnectionPool when sending data to GRPC
    """

    K8S_PORT_USES_AFTER = 8079
    K8S_PORT_USES_BEFORE = 8078
    K8S_PORT = 8080
    K8S_PORT_MONITORING = 9090

    def __init__(
            self,
            runtime_name,
            logger: Optional[JinaLogger] = None,
            compression: Optional[str] = None,
            metrics_registry: Optional['CollectorRegistry'] = None,
            meter: Optional['Meter'] = None,
            aio_tracing_client_interceptors: Optional[Sequence['ClientInterceptor']] = None,
            tracing_client_interceptor: Optional['OpenTelemetryClientInterceptor'] = None,
            channel_options: Optional[list] = None,
    ):
        self._logger = logger or JinaLogger(self.__class__.__name__)
        self.channel_options = channel_options

        self.compression = (
            getattr(grpc.Compression, compression)
            if compression
            else grpc.Compression.NoCompression
        )

        if metrics_registry:
            with ImportExtensions(
                    required=True,
                    help_text='You need to install the `prometheus_client` to use the montitoring functionality of jina',
            ):
                from prometheus_client import Summary

            sending_requests_time_metrics = Summary(
                'sending_request_seconds',
                'Time spent between sending a request to the Executor/Head and receiving the response',
                registry=metrics_registry,
                namespace='jina',
                labelnames=('runtime_name',),
            ).labels(runtime_name)

            received_response_bytes = Summary(
                'received_response_bytes',
                'Size in bytes of the response returned from the Head/Executor',
                registry=metrics_registry,
                namespace='jina',
                labelnames=('runtime_name',),
            ).labels(runtime_name)

            send_requests_bytes_metrics = Summary(
                'sent_request_bytes',
                'Size in bytes of the request sent to the Head/Executor',
                registry=metrics_registry,
                namespace='jina',
                labelnames=('runtime_name',),
            ).labels(runtime_name)
        else:
            sending_requests_time_metrics = None
            received_response_bytes = None
            send_requests_bytes_metrics = None

        self._metrics = _NetworkingMetrics(
            sending_requests_time_metrics,
            received_response_bytes,
            send_requests_bytes_metrics,
        )

        if meter:
            self._histograms = _NetworkingHistograms(
                sending_requests_time_metrics=meter.create_histogram(
                    name='jina_sending_request_seconds',
                    unit='s',
                    description='Time spent between sending a request to the Executor/Head and receiving the response',
                ),
                received_response_bytes=meter.create_histogram(
                    name='jina_received_response_bytes',
                    unit='By',
                    description='Size in bytes of the response returned from the Head/Executor',
                ),
                send_requests_bytes_metrics=meter.create_histogram(
                    name='jina_sent_request_bytes',
                    unit='By',
                    description='Size in bytes of the request sent to the Head/Executor',
                ),
                histogram_metric_labels={'runtime_name': runtime_name},
            )
        else:
            self._histograms = _NetworkingHistograms()

        self.aio_tracing_client_interceptors = aio_tracing_client_interceptors
        self.tracing_client_interceptor = tracing_client_interceptor
        self._connections = _ConnectionPoolMap(
            runtime_name=runtime_name,
            logger=self._logger,
            metrics=self._metrics,
            histograms=self._histograms,
            aio_tracing_client_interceptors=self.aio_tracing_client_interceptors,
            tracing_client_interceptor=self.tracing_client_interceptor,
            channel_options=self.channel_options,
        )
        self._deployment_address_map = {}

    def send_requests(
            self,
            requests: List[Request],
            deployment: str,
            head: bool = False,
            shard_id: Optional[int] = None,
            polling_type: PollingType = PollingType.ANY,
            endpoint: Optional[str] = None,
            metadata: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ) -> List[asyncio.Task]:
        """Send a request to target via one or all of the pooled connections, depending on polling_type

        :param requests: request (DataRequest) to send
        :param deployment: name of the Jina deployment to send the request to
        :param head: If True it is send to the head, otherwise to the worker pods
        :param shard_id: Send to a specific shard of the deployment, ignored for polling ALL
        :param polling_type: defines if the request should be send to any or all pooled connections for the target
        :param endpoint: endpoint to target with the requests
        :param metadata: metadata to send with the requests
        :param timeout: timeout for sending the requests
        :param retries: number of retries per gRPC call. If <0 it defaults to max(3, num_replicas)
        :return: list of asyncio.Task items for each send call
        """
        results = []
        connections = []
        if polling_type == PollingType.ANY:
            replica_list = self._connections.get_replicas(deployment, head, shard_id)
            if replica_list:
                connections.append(replica_list)
        elif polling_type == PollingType.ALL:
            shard_replica_lists = self._connections.get_replicas_all_shards(deployment)
            for replica_list in shard_replica_lists:
                connections.append(replica_list)
        else:
            raise ValueError(f'Unsupported polling type {polling_type}')

        for replica_list in connections:
            task = self._send_requests(
                requests,
                replica_list,
                endpoint=endpoint,
                metadata=metadata,
                timeout=timeout,
                retries=retries,
            )
            results.append(task)

        return results

    def send_discover_endpoint(
            self,
            deployment: str,
            head: bool = True,
            shard_id: Optional[int] = None,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ):
        """Sends a discover Endpoint call to target.

        :param deployment: name of the Jina deployment to send the request to
        :param head: If True it is send to the head, otherwise to the worker pods
        :param shard_id: Send to a specific shard of the deployment, ignored for polling ALL
        :param timeout: timeout for sending the requests
        :param retries: number of retries per gRPC call. If <0 it defaults to max(3, num_replicas)
        :return: coroutine items to send call
        """
        connection_list = self._connections.get_replicas(
            deployment, head, shard_id, True
        )
        if connection_list:
            return self._send_discover_endpoint(
                timeout=timeout, connection_list=connection_list, retries=retries
            )
        else:
            self._logger.debug(
                f'no available connections for deployment {deployment} and shard {shard_id}'
            )
            return None

    def send_requests_once(
            self,
            requests: List[Request],
            deployment: str,
            metadata: Optional[Dict[str, str]] = None,
            head: bool = False,
            shard_id: Optional[int] = None,
            endpoint: Optional[str] = None,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ) -> Optional[asyncio.Task]:
        """Send a request to target via only one of the pooled connections

        :param requests: request to send
        :param deployment: name of the Jina deployment to send the request to
        :param metadata: metadata to send with the request
        :param head: If True it is send to the head, otherwise to the worker pods
        :param shard_id: Send to a specific shard of the deployment, ignored for polling ALL
        :param endpoint: endpoint to target with the requests
        :param timeout: timeout for sending the requests
        :param retries: number of retries per gRPC call. If <0 it defaults to max(3, num_replicas)
        :return: asyncio.Task representing the send call
        """
        replicas = self._connections.get_replicas(deployment, head, shard_id)
        if replicas:
            result = self._send_requests(
                requests,
                replicas,
                endpoint=endpoint,
                metadata=metadata,
                timeout=timeout,
                retries=retries,
            )
            return result
        else:
            self._logger.debug(
                f'no available connections for deployment {deployment} and shard {shard_id}'
            )
            return None

    def send_single_document_request(
            self,
            request: SingleDocumentRequest,
            deployment: str,
            metadata: Optional[Dict[str, str]] = None,
            head: bool = False,
            endpoint: Optional[str] = None,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ) -> Optional[AsyncGenerator]:
        """Send a request to target via only one of the pooled connections

        :param request: request to send
        :param deployment: name of the Jina deployment to send the request to
        :param metadata: metadata to send with the request
        :param head: If True it is send to the head, otherwise to the worker pods
        :param endpoint: endpoint to target with the requests
        :param timeout: timeout for sending the requests
        :param retries: number of retries per gRPC call. If <0 it defaults to max(3, num_replicas)
        :return: asyncio.Task representing the send call
        """
        replicas = self._connections.get_replicas(deployment, head)
        if replicas:
            result_async_generator = self._send_single_doc_request(
                request,
                replicas,
                endpoint=endpoint,
                metadata=metadata,
                timeout=timeout,
                retries=retries,
            )
            return result_async_generator
        else:
            self._logger.debug(
                f'no available connections for deployment {deployment}'
            )
            return None

    def add_connection(
            self,
            deployment: str,
            address: str,
            head: bool = False,
            shard_id: Optional[int] = None,
    ):
        """
        Adds a connection for a deployment to this connection pool

        :param deployment: The deployment the connection belongs to, like 'encoder'
        :param head: True if the connection is for a head
        :param address: Address used for the grpc connection, format is <host>:<port>
        :param shard_id: Optional parameter to indicate this connection belongs to a shard, ignored for heads
        """
        if head:
            self._connections.add_head(deployment, address, 0)
        else:
            if shard_id is None:
                shard_id = 0
            self._connections.add_replica(deployment, shard_id, address)
        self._deployment_address_map[deployment] = address

    async def remove_connection(
            self,
            deployment: str,
            address: str,
            head: bool = False,
            shard_id: Optional[int] = None,
    ):
        """
        Removes a connection to a deployment

        :param deployment: The deployment the connection belongs to, like 'encoder'
        :param address: Address used for the grpc connection, format is <host>:<port>
        :param head: True if the connection is for a head
        :param shard_id: Optional parameter to indicate this connection belongs to a shard, ignored for heads
        :return: The removed connection, None if it did not exist
        """
        if head:
            return await self._connections.remove_head(deployment, address)
        else:
            if shard_id is None:
                shard_id = 0
            return await self._connections.remove_replica(deployment, address, shard_id)

    async def close(self):
        """
        Closes the connection pool
        """
        await self._connections.close()

    async def _handle_aiorpcerror(
            self,
            error: AioRpcError,
            retry_i: int = 0,
            request_id: str = '',
            tried_addresses: Set[str] = {
                ''
            },  # same deployment can have multiple addresses (replicas)
            total_num_tries: int = 1,  # number of retries + 1
            current_address: str = '',  # the specific address that was contacted during this attempt
            current_deployment: str = '',  # the specific deployment that was contacted during this attempt
            connection_list: Optional[_ReplicaList] = None,
            task_type: str = 'DataRequest'
    ) -> 'Optional[Union[AioRpcError, InternalNetworkError]]':
        # connection failures, cancelled requests, and timed out requests should be retried
        # all other cases should not be retried and will be raised immediately
        # connection failures have the code grpc.StatusCode.UNAVAILABLE
        # cancelled requests have the code grpc.StatusCode.CANCELLED
        # timed out requests have the code grpc.StatusCode.DEADLINE_EXCEEDED
        # if an Executor is down behind an API gateway, grpc.StatusCode.NOT_FOUND is returned
        # requests usually gets cancelled when the server shuts down
        # retries for cancelled requests will hit another replica in K8s
        skip_resetting = False
        if error.code() == grpc.StatusCode.UNAVAILABLE and 'not the leader' in error.details():
            self._logger.debug(f'RAFT node of {current_deployment} is not the leader. Trying next replica, if available.')
            skip_resetting = True # no need to reset, no problem with channel
        else:
            self._logger.debug(
                f'gRPC call to {current_deployment} for {task_type} errored, with error {format_grpc_error(error)} and for the {retry_i + 1}th time.'
            )
        errors_to_retry = [
            grpc.StatusCode.UNAVAILABLE,
            grpc.StatusCode.DEADLINE_EXCEEDED,
            grpc.StatusCode.NOT_FOUND,
        ]
        errors_to_handle = errors_to_retry + [
            grpc.StatusCode.CANCELLED,
            grpc.StatusCode.UNKNOWN,
            grpc.StatusCode.INTERNAL,
        ]

        if error.code() not in errors_to_handle:
            return error
        elif error.code() in errors_to_retry and retry_i >= total_num_tries - 1:
            self._logger.debug(
                f'gRPC call for {current_deployment} failed, retries exhausted'
            )
            from jina.excepts import InternalNetworkError

            # after connection failure the gRPC `channel` gets stuck in a failure state for a few seconds
            # removing and re-adding the connection (stub) is faster & more reliable than just waiting
            if connection_list and not skip_resetting:
                await connection_list.reset_connection(
                    current_address, current_deployment
                )

            return InternalNetworkError(
                og_exception=error,
                request_id=request_id,
                dest_addr=tried_addresses,
                details=error.details(),
            )
        else:
            if connection_list and not skip_resetting:
                await connection_list.reset_connection(
                    current_address, current_deployment
                )
            return None

    def _send_single_doc_request(
            self,
            request: SingleDocumentRequest,
            connections: _ReplicaList,
            endpoint: Optional[str] = None,
            metadata: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ) -> 'asyncio.Task[Union[Tuple, AioRpcError, InternalNetworkError]]':
        # this wraps the awaitable object from grpc as a coroutine so it can be used as a task
        # the grpc call function is not a coroutine but some _AioCall

        if endpoint:
            metadata = metadata or {}
            metadata['endpoint'] = endpoint

        if metadata:
            metadata = tuple(metadata.items())

        async def async_generator_wrapper():
            tried_addresses = set()
            num_replicas = len(connections.get_all_connections())
            if retries is None or retries < 0:
                total_num_tries = (
                        max(DEFAULT_MINIMUM_RETRIES, len(connections.get_all_connections()))
                        + 1
                )
            else:
                total_num_tries = 1 + retries  # try once, then do all the retries
            for i in range(total_num_tries):
                current_connection = None
                while (
                        current_connection is None
                        or current_connection.address in tried_addresses
                ):
                    current_connection = await connections.get_next_connection(
                        num_retries=total_num_tries
                    )
                    # if you request to retry more than the amount of replicas, we just skip, we could balance the
                    # retries in the future
                    if len(tried_addresses) >= num_replicas:
                        break
                tried_addresses.add(current_connection.address)
                try:
                    async for resp, metadata_resp in current_connection.send_single_doc_request(
                            request=request,
                            metadata=metadata,
                            compression=self.compression,
                            timeout=timeout,
                    ):
                        yield resp, metadata_resp
                    return
                except AioRpcError as e:
                    error = await self._handle_aiorpcerror(
                        error=e,
                        retry_i=i,
                        request_id=request.request_id,
                        tried_addresses=tried_addresses,
                        total_num_tries=total_num_tries,
                        current_address=current_connection.address,
                        current_deployment=current_connection.deployment_name,
                        connection_list=connections,
                        task_type='SingleDocumentRequest'
                    )
                    if error:
                        yield error, None
                        return
                except Exception as e:
                    yield e, None
                    return

        return async_generator_wrapper()

    def _send_requests(
            self,
            requests: List[Request],
            connections: _ReplicaList,
            endpoint: Optional[str] = None,
            metadata: Optional[Dict[str, str]] = None,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ) -> 'asyncio.Task[Union[Tuple, AioRpcError, InternalNetworkError]]':
        # this wraps the awaitable object from grpc as a coroutine so it can be used as a task
        # the grpc call function is not a coroutine but some _AioCall

        if endpoint:
            metadata = metadata or {}
            metadata['endpoint'] = endpoint

        if metadata:
            metadata = tuple(metadata.items())

        async def task_wrapper():
            tried_addresses = set()
            num_replicas = len(connections.get_all_connections())
            if retries is None or retries < 0:
                total_num_tries = (
                        max(DEFAULT_MINIMUM_RETRIES, len(connections.get_all_connections()))
                        + 1
                )
            else:
                total_num_tries = 1 + retries  # try once, then do all the retries
            for i in range(total_num_tries):
                current_connection = None
                while (
                        current_connection is None
                        or current_connection.address in tried_addresses
                ):
                    current_connection = await connections.get_next_connection(
                        num_retries=total_num_tries
                    )
                    # if you request to retry more than the amount of replicas, we just skip, we could balance the
                    # retries in the future
                    if len(tried_addresses) >= num_replicas:
                        break
                tried_addresses.add(current_connection.address)
                try:
                    return await current_connection.send_requests(
                        requests=requests,
                        metadata=metadata,
                        compression=self.compression,
                        timeout=timeout,
                    )
                except AioRpcError as e:
                    error = await self._handle_aiorpcerror(
                        error=e,
                        retry_i=i,
                        request_id=requests[0].request_id,
                        tried_addresses=tried_addresses,
                        total_num_tries=total_num_tries,
                        current_address=current_connection.address,
                        current_deployment=current_connection.deployment_name,
                        connection_list=connections,
                        task_type='DataRequest'
                    )
                    if error:
                        return error
                except Exception as e:
                    return e

        return asyncio.create_task(task_wrapper())

    def _send_discover_endpoint(
            self,
            connection_list: _ReplicaList,
            timeout: Optional[float] = None,
            retries: Optional[int] = -1,
    ):
        # this wraps the awaitable object from grpc as a coroutine so it can be used as a task
        # the grpc call function is not a coroutine but some _AioCall
        async def task_coroutine():
            tried_addresses = set()
            if retries is None or retries < 0:
                total_num_tries = (
                        max(
                            DEFAULT_MINIMUM_RETRIES,
                            len(connection_list.get_all_connections()),
                        )
                        + 1
                )
            else:
                total_num_tries = 1 + retries  # try once, then do all the retries
            for i in range(total_num_tries):
                connection = await connection_list.get_next_connection(
                    num_retries=total_num_tries
                )
                tried_addresses.add(connection.address)
                try:
                    return await connection.send_discover_endpoint(
                        timeout=timeout,
                    )
                except AioRpcError as e:
                    error = await self._handle_aiorpcerror(
                        error=e,
                        retry_i=i,
                        tried_addresses=tried_addresses,
                        current_address=connection.address,
                        current_deployment=connection.deployment_name,
                        connection_list=connection_list,
                        total_num_tries=total_num_tries,
                        task_type='EndpointDiscovery'
                    )
                    if error:
                        raise error
                except AttributeError:
                    return default_endpoints_proto, None

        return task_coroutine()

    def _get_all_replicas(self, deployment):
        replica_set = set()
        replica_set.update(self._connections.get_replicas_all_shards(deployment))
        replica_set.add(
            self._connections.get_replicas(deployment=deployment, head=True)
        )
        return set(filter(None, replica_set))



from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Tuple, Union

import grpc

from jina.proto import jina_pb2, jina_pb2_grpc
from jina.serve.instrumentation import MetricsTimer
from jina.serve.networking.instrumentation import (
    _NetworkingHistograms,
    _NetworkingMetrics,
)
from jina.serve.networking.utils import get_available_services, get_grpc_channel
from jina.types.request import Request
from jina.types.request.data import DataRequest, SingleDocumentRequest

if TYPE_CHECKING:  # pragma: no cover
    from grpc.aio._interceptor import ClientInterceptor


class _ConnectionStubs:
    """
    Maintains a list of grpc stubs available for a particular connection
    """

    STUB_MAPPING = {
        'jina.JinaDataRequestRPC': jina_pb2_grpc.JinaDataRequestRPCStub,
        'jina.JinaSingleDataRequestRPC': jina_pb2_grpc.JinaSingleDataRequestRPCStub,
        'jina.JinaSingleDocumentRequestRPC': jina_pb2_grpc.JinaSingleDocumentRequestRPCStub,
        'jina.JinaDiscoverEndpointsRPC': jina_pb2_grpc.JinaDiscoverEndpointsRPCStub,
        'jina.JinaRPC': jina_pb2_grpc.JinaRPCStub,
        'jina.JinaInfoRPC': jina_pb2_grpc.JinaInfoRPCStub,
    }

    def __init__(
        self,
        address,
        channel,
        deployment_name: str,
        metrics: _NetworkingMetrics,
        histograms: _NetworkingHistograms,
    ):
        self.address = address
        self.channel = channel
        self.deployment_name = deployment_name
        self._metrics = metrics
        self._histograms = histograms
        self._initialized = False

        if self._histograms:
            self.stub_specific_labels = {
                'deployment': deployment_name,
                'address': address,
            }

    # This has to be done lazily, because the target endpoint may not be available
    # when a connection is added
    async def _init_stubs(self):
        available_services = await get_available_services(self.channel)
        stubs = defaultdict(lambda: None)
        for service in available_services:
            if service in self.STUB_MAPPING:
                stubs[service] = self.STUB_MAPPING[service](self.channel)

        self.data_list_stub = stubs['jina.JinaDataRequestRPC']
        self.single_data_stub = stubs['jina.JinaSingleDataRequestRPC']
        self.stream_doc_stub = stubs['jina.JinaSingleDocumentRequestRPC']
        self.stream_stub = stubs['jina.JinaRPC']
        self.endpoints_discovery_stub = stubs['jina.JinaDiscoverEndpointsRPC']
        self.info_rpc_stub = stubs['jina.JinaInfoRPC']
        self._initialized = True

    async def send_discover_endpoint(
        self,
        timeout: Optional[float] = None,
    ) -> Tuple:
        """
        Use the endpoint discovery stub to request for the Endpoints Exposed by an Executor

        :param timeout: defines timeout for sending request

        :returns: Tuple of response and metadata about the response
        """
        if not self._initialized:
            await self._init_stubs()

        call_result = self.endpoints_discovery_stub.endpoint_discovery(
            jina_pb2.google_dot_protobuf_dot_empty__pb2.Empty(),
            timeout=timeout,
        )
        metadata, response = (
            await call_result.trailing_metadata(),
            await call_result,
        )
        return response, metadata

    def _get_metric_timer(self):
        if self._histograms.histogram_metric_labels is None:
            labels = None
        else:
            labels = {
                **self._histograms.histogram_metric_labels,
                **self.stub_specific_labels,
            }

        return MetricsTimer(
            self._metrics.sending_requests_time_metrics,
            self._histograms.sending_requests_time_metrics,
            labels,
        )

    def _record_request_bytes_metric(self, nbytes: int):
        if self._metrics.send_requests_bytes_metrics:
            self._metrics.send_requests_bytes_metrics.observe(nbytes)
        self._histograms.record_send_requests_bytes_metrics(
            nbytes, self.stub_specific_labels
        )

    def _record_received_bytes_metric(self, nbytes: int):
        if self._metrics.received_response_bytes:
            self._metrics.received_response_bytes.observe(nbytes)
        self._histograms.record_received_response_bytes(
            nbytes, self.stub_specific_labels
        )

    async def send_single_doc_request(self,
                                      request: SingleDocumentRequest,
                                      metadata,
                                      compression,
                                      timeout: Optional[float] = None):
        """
        Send requests and uses the appropriate grpc stub for this
        Stub is chosen based on availability and type of requests

        :param request: the requests to send
        :param metadata: the metadata to send alongside the requests
        :param compression: defines if compression should be used
        :param timeout: defines timeout for sending request

        :yields: Tuple of response and metadata about the response
        """
        if not self._initialized:
            await self._init_stubs()

        timer = self._get_metric_timer()
        if self.stream_doc_stub:
            self._record_request_bytes_metric(request.nbytes)

            with timer:
                async for response in self.stream_doc_stub.stream_doc(
                        request,
                        compression=compression,
                        timeout=timeout,
                        metadata=metadata,
                ):
                    self._record_received_bytes_metric(response.nbytes)
                    yield response, None
        else:
            raise ValueError(
                'Can not send SingleDocumentRequest. gRPC endpoint not available.'
            )

    async def send_requests(
        self,
        requests: List[Request],
        metadata,
        compression,
        timeout: Optional[float] = None,
    ) -> Tuple:
        """
        Send requests and uses the appropriate grpc stub for this
        Stub is chosen based on availability and type of requests

        :param requests: the requests to send
        :param metadata: the metadata to send alongside the requests
        :param compression: defines if compression should be used
        :param timeout: defines timeout for sending request

        :returns: Tuple of response and metadata about the response
        """
        if not self._initialized:
            await self._init_stubs()
        request_type = type(requests[0])

        timer = self._get_metric_timer()
        if request_type == DataRequest and len(requests) == 1:
            request = requests[0]
            if self.single_data_stub:
                self._record_request_bytes_metric(request.nbytes)
                call_result = self.single_data_stub.process_single_data(
                    request,
                    metadata=metadata,
                    compression=compression,
                    timeout=timeout,
                )
                with timer:
                    metadata, response = (
                        await call_result.trailing_metadata(),
                        await call_result,
                    )
                    self._record_received_bytes_metric(response.nbytes)
                return response, metadata

            elif self.stream_stub:
                self._record_request_bytes_metric(request.nbytes)

                with timer:
                    async for response in self.stream_stub.Call(
                        iter(requests),
                        compression=compression,
                        timeout=timeout,
                        metadata=metadata,
                    ):
                        self._record_received_bytes_metric(response.nbytes)
                        return response, None

        if request_type == DataRequest and len(requests) > 1:
            if self.data_list_stub:
                for request in requests:
                    self._record_request_bytes_metric(request.nbytes)
                call_result = self.data_list_stub.process_data(
                    requests,
                    metadata=metadata,
                    compression=compression,
                    timeout=timeout,
                )
                with timer:
                    metadata, response = (
                        await call_result.trailing_metadata(),
                        await call_result,
                    )
                    self._record_received_bytes_metric(response.nbytes)
                return response, metadata
            else:
                raise ValueError(
                    'Can not send list of DataRequests. gRPC endpoint not available.'
                )
        else:
            raise ValueError(f'Unsupported request type {type(requests[0])}')

    async def send_info_rpc(self, timeout: Optional[float] = None):
        """
        Use the JinaInfoRPC stub to send request to the _status endpoint exposed by the Runtime
        :param timeout: defines timeout for sending request
        :returns: JinaInfoProto
        """
        if not self._initialized:
            await self._init_stubs()

        call_result = self.info_rpc_stub._status(
            jina_pb2.google_dot_protobuf_dot_empty__pb2.Empty(),
            timeout=timeout,
        )
        return await call_result


def create_async_channel_stub(
    address,
    deployment_name: str,
    metrics: _NetworkingMetrics,
    histograms: _NetworkingHistograms,
    tls=False,
    root_certificates: Optional[str] = None,
    aio_tracing_client_interceptors: Optional[Sequence['ClientInterceptor']] = None,
    channel_options: Optional[Union[list, Dict[str, Any]]] = None,
) -> Tuple[_ConnectionStubs, grpc.aio.Channel]:
    """
    Creates an async GRPC Channel. This channel has to be closed eventually!

    :param address: the address to create the connection to, like 126.0.0.0.1:8080
    :param deployment_name: the name of the deployment (e.g. executor-1)
    :param tls: if True, use tls for the grpc channel
    :param root_certificates: the path to the root certificates for tls, only u
    :param metrics: NetworkingMetrics object that contain optional metrics
    :param histograms: NetworkingHistograms object that optionally record metrics
    :param aio_tracing_client_interceptors: List of async io gprc client tracing interceptors for tracing requests for asycnio channel
    :param channel_options: gRPC channel options
    :returns: DataRequest stubs and an async grpc channel
    """
    channel = get_grpc_channel(
        address,
        asyncio=True,
        tls=tls,
        root_certificates=root_certificates,
        aio_tracing_client_interceptors=aio_tracing_client_interceptors,
        options=channel_options,
    )

    return (
        _ConnectionStubs(address, channel, deployment_name, metrics, histograms),
        channel,
    )



import grpc
from jina.serve.consensus.add_voter.add_voter_pb2_grpc import RaftAdminStub
from jina.serve.consensus.add_voter.add_voter_pb2 import AddVoterRequest


def call_add_voter(target, replica_id, voter_address):
    with grpc.insecure_channel(target) as channel:
        stub = RaftAdminStub(channel)

        req = AddVoterRequest(
            id=replica_id,
            address=voter_address,
            previous_index=0,
        )

        try:
            future = stub.AddVoter(req)
            add_voter_result = stub.Await(future)
            _ = stub.Forget(future)
            if not add_voter_result.error:
                return True
            else:
                return False
        except:
            return False


async def async_call_add_voter(target, replica_id, voter_address):
    async with grpc.aio.insecure_channel(target) as channel:
        stub = RaftAdminStub(channel)

        req = AddVoterRequest(
            id=replica_id,
            address=voter_address,
            previous_index=0,
        )

        try:
            future = await stub.AddVoter(req)
            add_voter_result = await stub.Await(future)
            _ = await stub.Forget(future)
            if not add_voter_result.error:
                return True
            else:
                return False
        except:
            return False



import asyncio
import copy
import json
import multiprocessing
import os
import platform
import re
import subprocess
import sys
import threading
import time
import warnings
from argparse import Namespace
from collections import defaultdict
from contextlib import ExitStack
from itertools import cycle
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Type, Union, overload

from hubble.executor.helper import replace_secret_of_hub_uri
from rich import print
from rich.panel import Panel

from jina.clients import Client
from jina.clients.mixin import PostMixin
from jina.constants import (
    __default_executor__,
    __default_grpc_gateway__,
    __default_host__,
    __docker_host__,
    __windows__,
)
from jina.enums import (
    DeploymentRoleType,
    PodRoleType,
    PollingType,
    ProtocolType,
    ProviderType,
)
from jina.helper import (
    ArgNamespace,
    parse_host_scheme,
    random_port,
    send_telemetry_event,
)
from jina.importer import ImportExtensions
from jina.jaml import JAMLCompatible
from jina.logging.logger import JinaLogger
from jina.orchestrate.deployments.install_requirements_helper import (
    _get_package_path_from_uses,
    install_package_dependencies,
)
from jina.orchestrate.orchestrator import BaseOrchestrator
from jina.orchestrate.pods.factory import PodFactory
from jina.parsers import set_deployment_parser, set_gateway_parser
from jina.parsers.helper import _update_gateway_args
from jina.serve.networking.utils import host_is_local, in_docker

WRAPPED_SLICE_BASE = r'\[[-\d:]+\]'

if TYPE_CHECKING:
    from jina.clients.base import BaseClient
    from jina.serve.executors import BaseExecutor


class DeploymentType(type(ExitStack), type(JAMLCompatible)):
    """Type of Deployment, metaclass of :class:`Deployment`"""

    pass


def _call_add_voters(leader, voters, replica_ids, logger):
    # this method needs to be run in multiprocess, importing jraft in main process
    # makes it impossible to do tests sequentially
    from jina.serve.consensus.add_voter.call_add_voter import call_add_voter

    logger.debug(f'Trying to add {len(replica_ids)} voters to leader {leader}')
    success_lists = []
    for voter_address, replica_id in zip(voters, replica_ids):
        logger.debug(
            f'Trying to add replica-{str(replica_id)} as voter with address {voter_address} to leader at {leader}'
        )
        success = False
        for i in range(5):
            logger.debug(f'Trying {i}th time')
            success = call_add_voter(leader, str(replica_id), voter_address)
            if success:
                logger.debug(f'Trying {i}th time succeeded')
                break
            else:
                logger.debug(f'Trying {i}th failed. Wait 2 seconds for next try')
                time.sleep(2.0)

        success_lists.append(success)
        if not success:
            logger.warning(
                f'Failed to add {str(replica_id)} as voter with address {voter_address} to leader at {leader}. This could be because {leader} is not the leader, '
                f'and maybe the cluster is restoring from a previous cluster state'
            )
        else:
            logger.success(
                f'Replica-{str(replica_id)} successfully added as voter with address {voter_address} to leader at {leader}'
            )
    logger.debug('Adding voters to leader finished')
    return all(success_lists)


async def _async_call_add_voters(leader, voters, replica_ids, logger):
    # this method needs to be run in multiprocess, importing jraft in main process
    # makes it impossible to do tests sequentially
    from jina.serve.consensus.add_voter.call_add_voter import async_call_add_voter

    logger.debug(f'Trying to add {len(replica_ids)} voters to leader {leader}')
    success_lists = []
    for voter_address, replica_id in zip(voters, replica_ids):
        logger.debug(
            f'Trying to add replica-{str(replica_id)} as voter with address {voter_address} to leader at {leader}'
        )
        success = False
        for i in range(5):
            logger.debug(f'Trying {i}th time')
            success = await async_call_add_voter(leader, str(replica_id), voter_address)
            if success:
                logger.debug(f'Trying {i}th time succeeded')
                break
            else:
                logger.debug(f'Trying {i}th failed. Wait 2 seconds for next try')
                time.sleep(2.0)
        success_lists.append(success)
        if not success:
            logger.warning(
                f'Failed to add {str(replica_id)} as voter with address {voter_address} to leader at {leader}. This could be because {leader} is not the leader, '
                f'and maybe the cluster is restoring from a previous cluster state'
            )
        else:
            logger.success(
                f'Replica-{str(replica_id)} successfully added as voter with address {voter_address} to leader at {leader}'
            )
    logger.debug('Adding voters to leader finished')
    return all(success_lists)


class Deployment(JAMLCompatible, PostMixin, BaseOrchestrator, metaclass=DeploymentType):
    """A Deployment is an immutable set of pods, which run in replicas. They share the same input and output socket.
    Internally, the pods can run with the process/thread backend. They can also be run in their own containers
    :param args: arguments parsed from the CLI
    :param needs: deployments names of preceding deployments, the output of these deployments go into the input of this deployment
    """

    class _ReplicaSet:
        def __init__(
            self,
            deployment_args: Namespace,
            args: List[Namespace],
            head_pod,
            name,
        ):
            self.deployment_args = copy.copy(deployment_args)
            self.args = args
            self.shard_id = args[0].shard_id
            self._pods = []
            self.head_pod = head_pod
            self.name = name
            logger_kwargs = vars(self.deployment_args)
            logger_kwargs.pop('name')
            self.logger = JinaLogger(context=self.name, name=self.name, **logger_kwargs)

        def _add_voter_to_leader(self):
            leader_address = f'{self._pods[0].runtime_ctrl_address}'
            voter_addresses = [pod.runtime_ctrl_address for pod in self._pods[1:]]
            replica_ids = [pod.args.replica_id for pod in self._pods[1:]]
            self.logger.debug('Starting process to call Add Voters')
            res = _call_add_voters(
                leader=leader_address,
                voters=voter_addresses,
                replica_ids=replica_ids,
                logger=self.logger,
            )
            if res:
                self.logger.debug('Finished adding voters')
            else:
                self.logger.error('Adding Voters did not finish successfully')

        async def _async_add_voter_to_leader(self):
            leader_address = f'{self._pods[0].runtime_ctrl_address}'
            voter_addresses = [pod.runtime_ctrl_address for pod in self._pods[1:]]
            replica_ids = [pod.args.replica_id for pod in self._pods[1:]]
            self.logger.debug('Starting process to call Add Voters')
            res = await _async_call_add_voters(
                leader=leader_address,
                voters=voter_addresses,
                replica_ids=replica_ids,
                logger=self.logger,
            )
            if res:
                self.logger.debug('Finished adding voters')
            else:
                self.logger.error('Adding Voters did not finish successfully')

        @property
        def is_ready(self):
            return all(p.is_ready.is_set() for p in self._pods)

        def clear_pods(self):
            self._pods.clear()

        @property
        def num_pods(self):
            return len(self._pods)

        def join(self):
            for pod in self._pods:
                pod.join()

        def wait_start_success(self):
            self.logger.debug('Waiting for ReplicaSet to start successfully')
            for pod in self._pods:
                pod.wait_start_success()
            # should this be done only when the cluster is started ?
            if self._pods[0].args.stateful:
                self._add_voter_to_leader()
            self.logger.debug('ReplicaSet started successfully')

        async def async_wait_start_success(self):
            self.logger.debug('Waiting for ReplicaSet to start successfully')
            await asyncio.gather(
                *[pod.async_wait_start_success() for pod in self._pods]
            )
            # should this be done only when the cluster is started ?
            if self._pods[0].args.stateful:
                await self._async_add_voter_to_leader()
            self.logger.debug('ReplicaSet started successfully')

        def __enter__(self):
            for _args in self.args:
                _args.noblock_on_start = True
                pod = PodFactory.build_pod(_args).start()
                self._pods.append(pod)
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            closing_exception = None
            for pod in self._pods:
                try:
                    pod.close()
                except Exception as exc:
                    if closing_exception is None:
                        closing_exception = exc
            if exc_val is None and closing_exception is not None:
                raise closing_exception

    # overload_inject_start_deployment
    @overload
    def __init__(
        self,
        *,
        allow_concurrent: Optional[bool] = False,
        compression: Optional[str] = None,
        connection_list: Optional[str] = None,
        cors: Optional[bool] = False,
        description: Optional[str] = None,
        disable_auto_volume: Optional[bool] = False,
        docker_kwargs: Optional[dict] = None,
        entrypoint: Optional[str] = None,
        env: Optional[dict] = None,
        exit_on_exceptions: Optional[List[str]] = [],
        external: Optional[bool] = False,
        floating: Optional[bool] = False,
        force_update: Optional[bool] = False,
        gpus: Optional[str] = None,
        grpc_channel_options: Optional[dict] = None,
        grpc_metadata: Optional[dict] = None,
        grpc_server_options: Optional[dict] = None,
        host: Optional[List[str]] = ['0.0.0.0'],
        install_requirements: Optional[bool] = False,
        log_config: Optional[str] = None,
        metrics: Optional[bool] = False,
        metrics_exporter_host: Optional[str] = None,
        metrics_exporter_port: Optional[int] = None,
        monitoring: Optional[bool] = False,
        name: Optional[str] = 'executor',
        native: Optional[bool] = False,
        no_reduce: Optional[bool] = False,
        output_array_type: Optional[str] = None,
        polling: Optional[str] = 'ANY',
        port: Optional[int] = None,
        port_monitoring: Optional[int] = None,
        prefer_platform: Optional[str] = None,
        protocol: Optional[Union[str, List[str]]] = ['GRPC'],
        provider: Optional[str] = ['NONE'],
        py_modules: Optional[List[str]] = None,
        quiet: Optional[bool] = False,
        quiet_error: Optional[bool] = False,
        raft_configuration: Optional[dict] = None,
        reload: Optional[bool] = False,
        replicas: Optional[int] = 1,
        retries: Optional[int] = -1,
        runtime_cls: Optional[str] = 'WorkerRuntime',
        shards: Optional[int] = 1,
        ssl_certfile: Optional[str] = None,
        ssl_keyfile: Optional[str] = None,
        stateful: Optional[bool] = False,
        timeout_ctrl: Optional[int] = 60,
        timeout_ready: Optional[int] = 600000,
        timeout_send: Optional[int] = None,
        title: Optional[str] = None,
        tls: Optional[bool] = False,
        traces_exporter_host: Optional[str] = None,
        traces_exporter_port: Optional[int] = None,
        tracing: Optional[bool] = False,
        uses: Optional[Union[str, Type['BaseExecutor'], dict]] = 'BaseExecutor',
        uses_after: Optional[Union[str, Type['BaseExecutor'], dict]] = None,
        uses_after_address: Optional[str] = None,
        uses_before: Optional[Union[str, Type['BaseExecutor'], dict]] = None,
        uses_before_address: Optional[str] = None,
        uses_dynamic_batching: Optional[dict] = None,
        uses_metas: Optional[dict] = None,
        uses_requests: Optional[dict] = None,
        uses_with: Optional[dict] = None,
        uvicorn_kwargs: Optional[dict] = None,
        volumes: Optional[List[str]] = None,
        when: Optional[dict] = None,
        workspace: Optional[str] = None,
        **kwargs,
    ):
        """Create a Deployment to serve or deploy and Executor or Gateway

        :param allow_concurrent: Allow concurrent requests to be processed by the Executor. This is only recommended if the Executor is thread-safe.
        :param compression: The compression mechanism used when sending requests from the Head to the WorkerRuntimes. For more details, check https://grpc.github.io/grpc/python/grpc.html#compression.
        :param connection_list: dictionary JSON with a list of connections to configure
        :param cors: If set, a CORS middleware is added to FastAPI frontend to allow cross-origin access.
        :param description: The description of this HTTP server. It will be used in automatics docs such as Swagger UI.
        :param disable_auto_volume: Do not automatically mount a volume for dockerized Executors.
        :param docker_kwargs: Dictionary of kwargs arguments that will be passed to Docker SDK when starting the docker '
          container.

          More details can be found in the Docker SDK docs:  https://docker-py.readthedocs.io/en/stable/
        :param entrypoint: The entrypoint command overrides the ENTRYPOINT in Docker image. when not set then the Docker image ENTRYPOINT takes effective.
        :param env: The map of environment variables that are available inside runtime
        :param exit_on_exceptions: List of exceptions that will cause the Executor to shut down.
        :param external: The Deployment will be considered an external Deployment that has been started independently from the Flow.This Deployment will not be context managed by the Flow.
        :param floating: If set, the current Pod/Deployment can not be further chained, and the next `.add()` will chain after the last Pod/Deployment not this current one.
        :param force_update: If set, always pull the latest Hub Executor bundle even it exists on local
        :param gpus: This argument allows dockerized Jina Executors to discover local gpu devices.

              Note,
              - To access all gpus, use `--gpus all`.
              - To access multiple gpus, e.g. make use of 2 gpus, use `--gpus 2`.
              - To access specified gpus based on device id, use `--gpus device=[YOUR-GPU-DEVICE-ID]`
              - To access specified gpus based on multiple device id, use `--gpus device=[YOUR-GPU-DEVICE-ID1],device=[YOUR-GPU-DEVICE-ID2]`
              - To specify more parameters, use `--gpus device=[YOUR-GPU-DEVICE-ID],runtime=nvidia,capabilities=display
        :param grpc_channel_options: Dictionary of kwargs arguments that will be passed to the grpc channel as options when creating a channel, example : {'grpc.max_send_message_length': -1}. When max_attempts > 1, the 'grpc.service_config' option will not be applicable.
        :param grpc_metadata: The metadata to be passed to the gRPC request.
        :param grpc_server_options: Dictionary of kwargs arguments that will be passed to the grpc server as options when starting the server, example : {'grpc.max_send_message_length': -1}
        :param host: The host of the Gateway, which the client should connect to, by default it is 0.0.0.0. In the case of an external Executor (`--external` or `external=True`) this can be a list of hosts.  Then, every resulting address will be considered as one replica of the Executor.
        :param install_requirements: If set, try to install `requirements.txt` from the local Executor if exists in the Executor folder. If using Hub, install `requirements.txt` in the Hub Executor bundle to local.
        :param log_config: The config name or the absolute path to the YAML config file of the logger used in this object.
        :param metrics: If set, the sdk implementation of the OpenTelemetry metrics will be available for default monitoring and custom measurements. Otherwise a no-op implementation will be provided.
        :param metrics_exporter_host: If tracing is enabled, this hostname will be used to configure the metrics exporter agent.
        :param metrics_exporter_port: If tracing is enabled, this port will be used to configure the metrics exporter agent.
        :param monitoring: If set, spawn an http server with a prometheus endpoint to expose metrics
        :param name: The name of this object.

              This will be used in the following places:
              - how you refer to this object in Python/YAML/CLI
              - visualization
              - log message header
              - ...

              When not given, then the default naming strategy will apply.
        :param native: If set, only native Executors is allowed, and the Executor is always run inside WorkerRuntime.
        :param no_reduce: Disable the built-in reduction mechanism. Set this if the reduction is to be handled by the Executor itself by operating on a `docs_matrix` or `docs_map`
        :param output_array_type: The type of array `tensor` and `embedding` will be serialized to.

          Supports the same types as `docarray.to_protobuf(.., ndarray_type=...)`, which can be found
          `here <https://docarray.jina.ai/fundamentals/document/serialization/#from-to-protobuf>`.
          Defaults to retaining whatever type is returned by the Executor.
        :param polling: The polling strategy of the Deployment and its endpoints (when `shards>1`).
              Can be defined for all endpoints of a Deployment or by endpoint.
              Define per Deployment:
              - ANY: only one (whoever is idle) Pod polls the message
              - ALL: all Pods poll the message (like a broadcast)
              Define per Endpoint:
              JSON dict, {endpoint: PollingType}
              {'/custom': 'ALL', '/search': 'ANY', '*': 'ANY'}
        :param port: The port for input data to bind to, default is a random port between [49152, 65535]. In the case of an external Executor (`--external` or `external=True`) this can be a list of ports. Then, every resulting address will be considered as one replica of the Executor.
        :param port_monitoring: The port on which the prometheus server is exposed, default is a random port between [49152, 65535]
        :param prefer_platform: The preferred target Docker platform. (e.g. "linux/amd64", "linux/arm64")
        :param protocol: Communication protocol of the server exposed by the Executor. This can be a single value or a list of protocols, depending on your chosen Gateway. Choose the convenient protocols from: ['GRPC', 'HTTP', 'WEBSOCKET'].
        :param provider: If set, Executor is translated to a custom container compatible with the chosen provider. Choose the convenient providers from: ['NONE', 'SAGEMAKER'].
        :param py_modules: The customized python modules need to be imported before loading the executor

          Note that the recommended way is to only import a single module - a simple python file, if your
          executor can be defined in a single file, or an ``__init__.py`` file if you have multiple files,
          which should be structured as a python package. For more details, please see the
          `Executor cookbook <https://docs.jina.ai/concepts/executor/executor-files/>`__
        :param quiet: If set, then no log will be emitted from this object.
        :param quiet_error: If set, then exception stack information will not be added to the log
        :param raft_configuration: Dictionary of kwargs arguments that will be passed to the RAFT node as configuration options when starting the RAFT node.
        :param reload: If set, the Executor will restart while serving if YAML configuration source or Executor modules are changed. If YAML configuration is changed, the whole deployment is reloaded and new processes will be restarted. If only Python modules of the Executor have changed, they will be reloaded to the interpreter without restarting process.
        :param replicas: The number of replicas in the deployment
        :param retries: Number of retries per gRPC call. If <0 it defaults to max(3, num_replicas)
        :param runtime_cls: The runtime class to run inside the Pod
        :param shards: The number of shards in the deployment running at the same time. For more details check https://docs.jina.ai/concepts/flow/create-flow/#complex-flow-topologies
        :param ssl_certfile: the path to the certificate file
        :param ssl_keyfile: the path to the key file
        :param stateful: If set, start consensus module to make sure write operations are properly replicated between all the replicas
        :param timeout_ctrl: The timeout in milliseconds of the control request, -1 for waiting forever
        :param timeout_ready: The timeout in milliseconds of a Pod waits for the runtime to be ready, -1 for waiting forever
        :param timeout_send: The timeout in milliseconds used when sending data requests to Executors, -1 means no timeout, disabled by default
        :param title: The title of this HTTP server. It will be used in automatics docs such as Swagger UI.
        :param tls: If set, connect to deployment using tls encryption
        :param traces_exporter_host: If tracing is enabled, this hostname will be used to configure the trace exporter agent.
        :param traces_exporter_port: If tracing is enabled, this port will be used to configure the trace exporter agent.
        :param tracing: If set, the sdk implementation of the OpenTelemetry tracer will be available and will be enabled for automatic tracing of requests and customer span creation. Otherwise a no-op implementation will be provided.
        :param uses: The config of the executor, it could be one of the followings:
                  * the string literal of an Executor class name
                  * an Executor YAML file (.yml, .yaml, .jaml)
                  * a Jina Hub Executor (must start with `jinahub://` or `jinahub+docker://`)
                  * a docker image (must start with `docker://`)
                  * the string literal of a YAML config (must start with `!` or `jtype: `)
                  * the string literal of a JSON config

                  When use it under Python, one can use the following values additionally:
                  - a Python dict that represents the config
                  - a text file stream has `.read()` interface
        :param uses_after: The executor attached after the Pods described by --uses, typically used for receiving from all shards, accepted type follows `--uses`. This argument only applies for sharded Deployments (shards > 1).
        :param uses_after_address: The address of the uses-before runtime
        :param uses_before: The executor attached before the Pods described by --uses, typically before sending to all shards, accepted type follows `--uses`. This argument only applies for sharded Deployments (shards > 1).
        :param uses_before_address: The address of the uses-before runtime
        :param uses_dynamic_batching: Dictionary of keyword arguments that will override the `dynamic_batching` configuration in `uses`
        :param uses_metas: Dictionary of keyword arguments that will override the `metas` configuration in `uses`
        :param uses_requests: Dictionary of keyword arguments that will override the `requests` configuration in `uses`
        :param uses_with: Dictionary of keyword arguments that will override the `with` configuration in `uses`
        :param uvicorn_kwargs: Dictionary of kwargs arguments that will be passed to Uvicorn server when starting the server

          More details can be found in Uvicorn docs: https://www.uvicorn.org/settings/
        :param volumes: The path on the host to be mounted inside the container.

          Note,
          - If separated by `:`, then the first part will be considered as the local host path and the second part is the path in the container system.
          - If no split provided, then the basename of that directory will be mounted into container's root path, e.g. `--volumes="/user/test/my-workspace"` will be mounted into `/my-workspace` inside the container.
          - All volumes are mounted with read-write mode.
        :param when: The condition that the documents need to fulfill before reaching the Executor.The condition can be defined in the form of a `DocArray query condition <https://docarray.jina.ai/fundamentals/documentarray/find/#query-by-conditions>`
        :param workspace: The working directory for any IO operations in this object. If not set, then derive from its parent `workspace`.

        .. # noqa: DAR202
        .. # noqa: DAR101
        .. # noqa: DAR003
        """

    # overload_inject_end_deployment

    def __init__(
        self,
        args: Union['Namespace', Dict, None] = None,
        needs: Optional[Set[str]] = None,
        include_gateway: bool = True,
        **kwargs,
    ):
        super().__init__()
        self._gateway_kwargs = {}
        self._include_gateway = include_gateway
        if self._include_gateway:
            # arguments exclusive to the gateway
            for field in ['port', 'ports']:
                if field in kwargs:
                    self._gateway_kwargs[field] = kwargs.pop(field)

            # arguments common to both gateway and the Executor
            for field in ['host', 'log_config']:
                if field in kwargs:
                    self._gateway_kwargs[field] = kwargs[field]

        parser = set_deployment_parser()
        if args is None:
            args = ArgNamespace.kwargs2namespace(kwargs, parser, True)
        self.args = args
        self._gateway_load_balancer = False
        if self.args.provider == ProviderType.SAGEMAKER:
            if self._gateway_kwargs.get('port', 0) == 8080:
                raise ValueError(
                    'Port 8080 is reserved for Sagemaker deployment. '
                    'Please use another port'
                )
            if self.args.port != [8080]:
                warnings.warn(
                    'Port is changed to 8080 for Sagemaker deployment. '
                    f'Port {self.args.port} is ignored'
                )
                self.args.port = [8080]
            if self.args.protocol != [ProtocolType.HTTP]:
                warnings.warn(
                    'Protocol is changed to HTTP for Sagemaker deployment. '
                    f'Protocol {self.args.protocol} is ignored'
                )
                self.args.protocol = [ProtocolType.HTTP]
        if self._include_gateway and ProtocolType.HTTP in self.args.protocol:
            self._gateway_load_balancer = True
        log_config = kwargs.get('log_config')
        if log_config:
            self.args.log_config = log_config
        self.args.polling = (
            args.polling if hasattr(args, 'polling') else PollingType.ANY
        )
        # polling only works for shards, if there are none, polling will be ignored
        if getattr(args, 'shards', 1) == 1:
            self.args.polling = PollingType.ANY

        if (
            getattr(args, 'shards', 1) > 1
            and ProtocolType.HTTP in self.args.protocol
            and self.args.deployment_role != DeploymentRoleType.GATEWAY
        ):
            raise RuntimeError(
                f'It is not supported to have {ProtocolType.HTTP.to_string()} deployment for '
                f'Deployments with more than one shard'
            )

        if (
            ProtocolType.WEBSOCKET in self.args.protocol
            and self.args.deployment_role != DeploymentRoleType.GATEWAY
        ):
            raise RuntimeError(
                f'It is not supported to have {ProtocolType.WEBSOCKET.to_string()} deployment for '
                f'Deployments'
            )
        is_mac_os = platform.system() == 'Darwin'
        is_windows_os = platform.system() == 'Windows'
        is_37 = sys.version_info.major == 3 and sys.version_info.minor == 7

        if self.args.stateful and (is_windows_os or (is_mac_os and is_37)):
            if is_windows_os:
                raise RuntimeError('Stateful feature is not available on Windows')
            if is_mac_os:
                raise RuntimeError(
                    'Stateful feature when running on MacOS requires Python3.8 or newer version'
                )
        if self.args.stateful and (
            ProtocolType.WEBSOCKET in self.args.protocol
            or ProtocolType.HTTP in self.args.protocol
            or len(self.args.protocol) > 1
        ):
            raise RuntimeError(
                f'Stateful feature is only available for Deployments using a single {ProtocolType.GRPC.to_string()} protocol. {self.args.protocol} were requested'
            )
        self.needs = (
            needs or set()
        )  #: used in the :class:`jina.flow.Flow` to build the graph

        # parse addresses for distributed replicas
        (
            self.ext_repl_hosts,
            self.ext_repl_ports,
            self.ext_repl_schemes,
            self.ext_repl_tls,
        ) = ([], [], [], [])
        if self.args.pod_role != PodRoleType.GATEWAY:
            self._parse_external_replica_hosts_and_ports()
            self._parse_addresses_into_host_and_port()
        if len(self.ext_repl_ports) > 1:
            if self.args.replicas != 1 and self.args.replicas != len(
                self.ext_repl_ports
            ):
                raise ValueError(
                    f'Number of hosts ({len(self.args.host)}) does not match the number of replicas ({self.args.replicas})'
                )
            elif self.args.external:
                self.args.replicas = len(self.ext_repl_ports)

        self.uses_before_pod = None
        self.uses_after_pod = None
        self.head_pod = None
        self.gateway_pod = None
        self.shards = {}
        self._update_port_monitoring_args()
        self.update_pod_args()

        if self._include_gateway:
            gateway_parser = set_gateway_parser()
            args = ArgNamespace.kwargs2namespace(
                self._gateway_kwargs, gateway_parser, True
            )
            args.protocol = self.args.protocol

            args.deployments_addresses = json.dumps(
                {'executor': self._get_connection_list_for_single_executor()}
            )
            args.graph_description = (
                '{"start-gateway": ["executor"], "executor": ["end-gateway"]}'
            )
            _update_gateway_args(
                args, gateway_load_balancer=self._gateway_load_balancer
            )
            self.pod_args['gateway'] = args
        else:
            self.pod_args['gateway'] = None

        self.logger = JinaLogger(self.__class__.__name__, **vars(self.args))

    def _get_connection_list_for_flow(self) -> List[str]:
        if self.head_args:
            # add head information
            return [f'{self.protocol.lower()}://{self.host}:{self.head_port}']
        else:
            # there is no head, add the worker connection information instead
            ports = self.ports
            hosts = [
                __docker_host__
                if host_is_local(host) and in_docker() and self._is_docker
                else host
                for host in self.hosts
            ]
            return [
                f'{self.protocol.lower()}://{host}:{port}'
                for host, port in zip(hosts, ports)
            ]

    def _get_connection_list_for_single_executor(
        self,
    ) -> Union[List[str], Dict[str, List[str]]]:
        if self.head_args:
            # add head information
            return [f'{self.protocol.lower()}://{self.host}:{self.head_port}']
        else:
            # there is no head, add the worker connection information instead
            ports_dict = defaultdict(list)
            for replica_pod_arg in self.pod_args['pods'][0]:
                for protocol, port in zip(
                    replica_pod_arg.protocol, replica_pod_arg.port
                ):
                    ports_dict[str(protocol)].append(port)

            host = (
                __docker_host__
                if host_is_local(self.args.host[0]) and in_docker() and self._is_docker
                else self.args.host[0]
            )

            connection_dict = {}
            for protocol, ports in ports_dict.items():
                connection_dict[protocol] = [
                    f'{protocol.lower()}://{host}:{port}' for port in ports
                ]
            return connection_dict

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        super().__exit__(exc_type, exc_val, exc_tb)
        self.join()
        if self._include_gateway:
            self._stop_time = time.time()
            send_telemetry_event(
                event='stop',
                obj_cls_name=self.__class__.__name__,
                entity_id=self._entity_id,
                duration=self._stop_time - self._start_time,
                exc_type=str(exc_type),
            )
        self.logger.close()

    def _parse_addresses_into_host_and_port(self):
        # splits addresses passed to `host` into separate `host` and `port`

        for i, _host in enumerate(self.args.host):
            _hostname, port, scheme, tls = parse_host_scheme(_host)
            if _hostname != _host:  # more than just hostname was passed to `host`
                self.args.host[i] = _hostname
                self.args.port[i] = port
                self.args.scheme = scheme
                self.args.tls = tls
        for i, repl_host in enumerate(self.ext_repl_hosts):
            _hostname, port, scheme, tls = parse_host_scheme(repl_host)
            if (
                _hostname != self.ext_repl_hosts[i]
            ):  # more than just hostname was passed to `host`
                self.ext_repl_hosts[i] = _hostname
                self.ext_repl_ports[i] = port
                self.ext_repl_schemes[i] = scheme
                self.ext_repl_tls[i] = tls

    def _parse_external_replica_hosts_and_ports(self):
        # splits user provided lists of hosts and ports into a host and port for every distributed replica
        ext_repl_ports: List = self.args.port.copy()
        ext_repl_hosts: List = self.args.host.copy()
        if len(ext_repl_hosts) < len(ext_repl_ports):
            if (
                len(ext_repl_hosts) == 1
            ):  # only one host given, assume replicas are on the same host
                ext_repl_hosts = ext_repl_hosts * len(ext_repl_ports)
                self.args.host = self.args.host * len(ext_repl_ports)
        elif len(ext_repl_hosts) > len(ext_repl_ports):
            if (
                len(ext_repl_ports) == 1
            ):  # only one port given, assume replicas are on the same port
                ext_repl_ports = ext_repl_ports * len(ext_repl_hosts)
                self.args.port = self.args.port * len(ext_repl_hosts)
        if len(ext_repl_hosts) != len(ext_repl_ports):
            raise ValueError(
                f'Number of hosts ({len(ext_repl_hosts)}) does not match the number of ports ({len(ext_repl_ports)})'
            )

        self.ext_repl_hosts, self.ext_repl_ports = ext_repl_hosts, ext_repl_ports
        # varying tls and schemes other than 'grpc' only implemented if the entire address is passed to `host`
        self.ext_repl_schemes = [
            getattr(self.args, 'scheme', None) for _ in self.ext_repl_ports
        ]
        self.ext_repl_tls = [
            getattr(self.args, 'tls', None) for _ in self.ext_repl_ports
        ]

    def _update_port_monitoring_args(self):
        _all_port_monitoring = self.args.port_monitoring
        self.args.all_port_monitoring = (
            [_all_port_monitoring]
            if not type(_all_port_monitoring) == list
            else _all_port_monitoring
        )
        self.args.port_monitoring = int(
            self.args.all_port_monitoring[0]
        )  # this is for the head

    def update_pod_args(self):
        """Update args of all its pods based on Deployment args. Including head/tail"""
        if self.args.runtime_cls == 'GatewayRuntime':
            _update_gateway_args(self.args)
        if isinstance(self.args, Dict):
            # This is used when a Deployment is created in a remote context, where pods & their connections are already given.
            self.pod_args = self.args
        else:
            self.pod_args = self._parse_args(self.args)

    def update_worker_pod_args(self):
        """Update args of all its worker pods based on Deployment args. Does not touch head and tail"""
        self.pod_args['pods'] = self._set_pod_args()

    @property
    def role(self) -> 'DeploymentRoleType':
        """Return the role of this :class:`Deployment`.

        .. # noqa: DAR201
        """
        return self.args.deployment_role

    @property
    def name(self) -> str:
        """The name of this :class:`Deployment`.


        .. # noqa: DAR201
        """
        return self.args.name

    @property
    def head_host(self) -> str:
        """Get the host of the HeadPod of this deployment
        .. # noqa: DAR201
        """
        return self.head_args.host if self.head_args else None

    @property
    def head_port(self):
        """Get the port of the HeadPod of this deployment
        .. # noqa: DAR201
        """
        return self.head_args.port[0] if self.head_args else None

    @property
    def head_port_monitoring(self):
        """Get the port_monitoring of the HeadPod of this deployment
        .. # noqa: DAR201
        """
        return self.head_args.port_monitoring if self.head_args else None

    @property
    def client(self) -> 'BaseClient':
        """Return a :class:`BaseClient` object attach to this Flow.

        .. # noqa: DAR201"""

        kwargs = dict(
            host=self.host,
            port=self.port,
            protocol=self.protocol,
            grpc_channel_options=self.args.grpc_channel_options,
        )
        kwargs.update(self._gateway_kwargs)
        return Client(**kwargs)

    @staticmethod
    def _copy_to_head_args(args: Namespace) -> Namespace:
        """
        Set the outgoing args of the head router

        :param args: basic arguments
        :return: enriched head arguments
        """

        _head_args = copy.deepcopy(args)
        _head_args.polling = args.polling
        _head_args.port = args.port
        _head_args.host = args.host[0]
        _head_args.uses = args.uses
        _head_args.pod_role = PodRoleType.HEAD
        _head_args.runtime_cls = 'HeadRuntime'
        _head_args.replicas = 1

        if args.name:
            _head_args.name = f'{args.name}/head'
        else:
            _head_args.name = 'head'

        return _head_args

    @property
    def deployments(self) -> List[Dict]:
        """Get deployments of the deployment. The Deployment just gives one deployment.

        :return: list of deployments
        """
        return [
            {
                'name': self.name,
                'head_host': self.head_host,
                'head_port': self.head_port,
            }
        ]

    @property
    def _is_docker(self) -> bool:
        """
        Check if this deployment is to be run in Docker.

        :return: True if this deployment is to be run in Docker
        """
        from hubble.executor.helper import is_valid_docker_uri

        uses = getattr(self.args, 'uses', '')
        return is_valid_docker_uri(uses)

    @property
    def _is_executor_from_yaml(self) -> bool:
        """
        Check if this deployment is to be run from YAML configuration.

        :return: True if this deployment is to be run from YAML configuration
        """
        uses = getattr(self.args, 'uses', '')
        return uses.endswith('yml') or uses.endswith('yaml')

    @property
    def tls_enabled(self):
        """
        Check if secure connection via TLS is enabled for this Deployment.

        :return: True if tls is enabled, False otherwise
        """
        has_cert = getattr(self.args, 'ssl_certfile', None) is not None
        has_key = getattr(self.args, 'ssl_keyfile', None) is not None
        tls = getattr(self.args, 'tls', False)
        return tls or (has_cert and has_key)

    @property
    def external(self) -> bool:
        """
        Check if this deployment is external.

        :return: True if this deployment is provided as an external deployment, False otherwise
        """
        return getattr(self.args, 'external', False)

    @property
    def grpc_metadata(self):
        """
        Get the gRPC metadata for this deployment
        :return: The gRPC metadata for this deployment. If the deployment is a Gateway, return None.
        """
        return getattr(self.args, 'grpc_metadata', None)

    @property
    def protocol(self):
        """
        :return: the protocol of this deployment
        """
        args = self.pod_args['gateway'] or self.args

        protocol = getattr(args, 'protocol', ['grpc'])
        if not isinstance(protocol, list):
            protocol = [protocol]
        protocol = [
            str(_p).lower() + ('s' if self.tls_enabled else '') for _p in protocol
        ]
        if len(protocol) == 1:
            return protocol[0]
        else:
            return protocol

    @property
    def first_pod_args(self) -> Namespace:
        """Return the first worker pod's args


        .. # noqa: DAR201
        """
        # note this will be never out of boundary
        return self.pod_args['gateway'] or self.pod_args['pods'][0][0]

    @property
    def host(self) -> str:
        """Get the host name of this deployment


        .. # noqa: DAR201
        """
        return self.first_pod_args.host

    @property
    def port(self):
        """
        :return: The port of this deployment
        """
        return self.first_pod_args.port[0]

    @property
    def ports(self) -> List[int]:
        """Returns a list of ports exposed by this Deployment.
        Exposed means these are the ports a Client/Gateway is supposed to communicate with.
        For sharded deployments this will be the head_port.
        For non-sharded deployments it will be all replica ports.
        .. # noqa: DAR201
        """
        if self.head_port:
            return [self.head_port]
        else:
            ports = []
            for replica in self.pod_args['pods'][0]:
                if isinstance(replica.port, list):
                    ports.extend(replica.port)
                else:
                    ports.append(replica.port)
            return ports

    @property
    def hosts(self) -> List[str]:
        """Returns a list of host addresses exposed by this Deployment.
        Exposed means these are the host a Client/Gateway is supposed to communicate with.
        For sharded deployments this will be the head host.
        For non-sharded deployments it will be all replica hosts.
        .. # noqa: DAR201
        """
        if self.head_host:
            return [self.head_host]
        else:
            return [replica.host for replica in self.pod_args['pods'][0]]

    def _parse_args(
        self, args: Namespace
    ) -> Dict[str, Optional[Union[List[Namespace], Namespace]]]:
        return self._parse_base_deployment_args(args)

    @property
    def head_args(self) -> Namespace:
        """Get the arguments for the `head` of this Deployment.


        .. # noqa: DAR201
        """
        return self.pod_args['head']

    @head_args.setter
    def head_args(self, args):
        """Set the arguments for the `head` of this Deployment.


        .. # noqa: DAR101
        """
        self.pod_args['head'] = args

    @property
    def uses_before_args(self) -> Namespace:
        """Get the arguments for the `uses_before` of this Deployment.


        .. # noqa: DAR201
        """
        return self.pod_args['uses_before']

    @uses_before_args.setter
    def uses_before_args(self, args):
        """Set the arguments for the `uses_before` of this Deployment.


        .. # noqa: DAR101
        """
        self.pod_args['uses_before'] = args

    @property
    def uses_after_args(self) -> Namespace:
        """Get the arguments for the `uses_after` of this Deployment.


        .. # noqa: DAR201
        """
        return self.pod_args['uses_after']

    @uses_after_args.setter
    def uses_after_args(self, args):
        """Set the arguments for the `uses_after` of this Deployment.


        .. # noqa: DAR101
        """
        self.pod_args['uses_after'] = args

    @property
    def all_args(self) -> List[Namespace]:
        """Get all arguments of all Pods in this Deployment.

        .. # noqa: DAR201
        """
        all_args = (
            ([self.pod_args['uses_before']] if self.pod_args['uses_before'] else [])
            + ([self.pod_args['uses_after']] if self.pod_args['uses_after'] else [])
            + ([self.pod_args['head']] if self.pod_args['head'] else [])
            + ([self.pod_args['gateway']] if self._include_gateway else [])
        )
        for shard_id in self.pod_args['pods']:
            all_args += self.pod_args['pods'][shard_id]
        return all_args

    @property
    def num_pods(self) -> int:
        """Get the number of running :class:`Pod`

        .. # noqa: DAR201
        """
        num_pods = 0
        if self.head_pod is not None:
            num_pods += 1
        if self.uses_before_pod is not None:
            num_pods += 1
        if self.uses_after_pod is not None:
            num_pods += 1
        if self.gateway_pod is not None:
            num_pods += 1
        if self.shards:  # external deployments
            for shard_id in self.shards:
                num_pods += self.shards[shard_id].num_pods
        return num_pods

    def __eq__(self, other: 'Deployment'):
        return self.num_pods == other.num_pods and self.name == other.name

    @staticmethod
    def get_worker_host(pod_args, pod_is_container, head_is_container):
        """
        Check if the current pod and head are both containerized on the same host
        If so, __docker_host__ needs to be advertised as the worker's address to the head

        :param pod_args: arguments of the worker pod
        :param pod_is_container: boolean specifying if pod is to be run in container
        :param head_is_container: boolean specifying if head pod is to be run in container
        :return: host to pass in connection list of the head
        """
        # Check if the current pod and head are both containerized on the same host
        # If so __docker_host__ needs to be advertised as the worker's address to the head
        worker_host = (
            __docker_host__
            if (pod_is_container and (head_is_container or in_docker()))
            and host_is_local(pod_args.host)
            else pod_args.host
        )
        return worker_host

    def _wait_until_all_ready(self):
        import warnings

        with warnings.catch_warnings():
            wait_for_ready_coro = self.async_wait_start_success()

            try:
                _ = asyncio.get_event_loop()
            except:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)

            async def _f():
                pass

            running_in_event_loop = False
            try:
                asyncio.get_event_loop().run_until_complete(_f())
            except:
                running_in_event_loop = True

            if not running_in_event_loop:
                asyncio.get_event_loop().run_until_complete(wait_for_ready_coro)
            else:
                wait_ready_thread = threading.Thread(
                    target=self.wait_start_success, daemon=True
                )
                wait_ready_thread.start()
                wait_ready_thread.join()

    def start(self) -> 'Deployment':
        """
        Start to run all :class:`Pod` in this Deployment.

        :return: started deployment

        .. note::
            If one of the :class:`Pod` fails to start, make sure that all of them
            are properly closed.
        """

        self._start_time = time.time()

        if not self._is_docker and getattr(self.args, 'install_requirements', False):
            install_package_dependencies(_get_package_path_from_uses(self.args.uses))

        if self.pod_args['uses_before'] is not None:
            _args = self.pod_args['uses_before']
            _args.noblock_on_start = True
            self.uses_before_pod = PodFactory.build_pod(_args)
            self.enter_context(self.uses_before_pod)
        if self.pod_args['uses_after'] is not None:
            _args = self.pod_args['uses_after']
            _args.noblock_on_start = True
            self.uses_after_pod = PodFactory.build_pod(_args)
            self.enter_context(self.uses_after_pod)

        num_shards = len(self.pod_args['pods'])
        for shard_id in self.pod_args['pods']:
            self.shards[shard_id] = self._ReplicaSet(
                deployment_args=self.args,
                args=self.pod_args['pods'][shard_id],
                head_pod=self.head_pod,
                name=f'{self.name}-replica-set-{shard_id}'
                if num_shards > 1
                else f'{self.name}-replica-set',
            )
            self.enter_context(self.shards[shard_id])

        if self.pod_args['head'] is not None:
            _args = self.pod_args['head']
            _args.noblock_on_start = True
            self.head_pod = PodFactory.build_pod(_args)
            self.enter_context(self.head_pod)

        if self._include_gateway:
            _args = self.pod_args['gateway']
            _args.noblock_on_start = True
            self.gateway_pod = PodFactory.build_pod(
                _args, gateway_load_balancer=self._gateway_load_balancer
            )
            self.enter_context(self.gateway_pod)

        if not self.args.noblock_on_start:
            self._wait_until_all_ready()
        if self._include_gateway:
            all_panels = []
            self._get_summary_table(all_panels)

            from rich.rule import Rule

            print(Rule(':tada: Deployment is ready to serve!'), *all_panels)

            send_telemetry_event(
                event='start',
                obj_cls_name=self.__class__.__name__,
                entity_id=self._entity_id,
            )

        return self

    def wait_start_success(self) -> None:
        """Block until all pods start successfully.

        If not successful, it will raise an error hoping the outer function will catch it
        """
        try:
            if self.uses_before_pod is not None:
                self.uses_before_pod.wait_start_success()
            if self.uses_after_pod is not None:
                self.uses_after_pod.wait_start_success()
            if self.head_pod is not None:
                self.head_pod.wait_start_success()
            if self.gateway_pod is not None:
                self.gateway_pod.wait_start_success()
            for shard_id in self.shards:
                self.shards[shard_id].wait_start_success()
        except:
            self.close()
            raise

    async def async_wait_start_success(self) -> None:
        """Block until all pods start successfully.

        If unsuccessful, it will raise an error hoping the outer function will catch it
        """
        try:
            coros = []
            if self.uses_before_pod is not None:
                coros.append(self.uses_before_pod.async_wait_start_success())
            if self.uses_after_pod is not None:
                coros.append(self.uses_after_pod.async_wait_start_success())
            if self.gateway_pod is not None:
                coros.append(self.gateway_pod.async_wait_start_success())
            if self.head_pod is not None:
                coros.append(self.head_pod.async_wait_start_success())
            for shard_id in self.shards:
                coros.append(self.shards[shard_id].async_wait_start_success())

            await asyncio.gather(*coros)
            self.logger.debug('Deployment started successfully')
        except:
            self.close()
            raise

    def join(self):
        """Wait until all pods exit"""
        try:
            if self.uses_before_pod is not None:
                self.uses_before_pod.join()
            if self.uses_after_pod is not None:
                self.uses_after_pod.join()
            if self.shards:
                for shard_id in self.shards:
                    self.shards[shard_id].join()
            if self.head_pod is not None:
                self.head_pod.join()
            if self.gateway_pod is not None:
                self.gateway_pod.join()
        except KeyboardInterrupt:
            pass
        finally:
            self.head_pod = None
            if self.shards:
                for shard_id in self.shards:
                    self.shards[shard_id].clear_pods()

    @property
    def is_ready(self) -> bool:
        """Check if Deployment is ready

        .. note::
            A Deployment is ready when all the Pods it contains are ready


        .. # noqa: DAR201
        """
        is_ready = True
        if self.head_pod is not None:
            is_ready = self.head_pod.is_ready.is_set()
        if is_ready:
            for shard_id in self.shards:
                is_ready = self.shards[shard_id].is_ready
        if is_ready and self.uses_before_pod is not None:
            is_ready = self.uses_before_pod.is_ready.is_set()
        if is_ready and self.uses_after_pod is not None:
            is_ready = self.uses_after_pod.is_ready.is_set()
        if is_ready and self.gateway_pod is not None:
            is_ready = self.gateway_pod.is_ready.is_set()
        return is_ready

    @staticmethod
    def _parse_devices(value: str, num_devices: int):
        """Parse a list of devices from string, like `start:stop:step` or 'num1,num2,num3` or combination of both.

        :param value: a string-like
        :param num_devices: total number of devices
        :return: slice
        """

        use_uuids = False
        if re.match(WRAPPED_SLICE_BASE, value):
            value = value[1:-1]

        if value:
            parts = value.split(',')
            if len(parts) == 1:
                parts = value.split(':')

                if len(parts) == 1:
                    try:
                        int(parts[0])
                    except:
                        use_uuids = True
                    if use_uuids:
                        return parts
                    parts = [parts[0], str(int(parts[0]) + 1)]
            else:
                # try to detect if parts are not numbers
                try:
                    int(parts[0])
                except:
                    use_uuids = True

                if not use_uuids:
                    return [int(p) for p in parts]
                else:
                    return parts
        else:
            parts = []

        all_devices = range(num_devices)
        return all_devices[slice(*[int(p) if p else None for p in parts])]

    @staticmethod
    def _roundrobin_cuda_device(device_str: Optional[str], replicas: int):
        """Parse CUDA device string with RR prefix

        :param device_str: `RRm:n`, where `RR` is the prefix, m:n is python slice format
        :param replicas: the number of replicas
        :return: a map from replica ID to device ID
        """
        if (
            device_str
            and isinstance(device_str, str)
            and device_str.startswith('RR')
            and replicas >= 1
        ):
            try:
                num_devices = str(subprocess.check_output(['nvidia-smi', '-L'])).count(
                    'UUID'
                )
            except:
                num_devices = int(os.environ.get('CUDA_TOTAL_DEVICES', 0))
                if num_devices == 0:
                    return

            selected_devices = []
            if device_str[2:]:
                for device in Deployment._parse_devices(device_str[2:], num_devices):
                    selected_devices.append(device)
            else:
                selected_devices = range(num_devices)
            _c = cycle(selected_devices)
            return {j: next(_c) for j in range(replicas)}

    def _set_pod_args(self) -> Dict[int, List[Namespace]]:
        result = {}
        shards = getattr(self.args, 'shards', 1)
        replicas = getattr(self.args, 'replicas', 1)
        if self.args.deployment_role == DeploymentRoleType.GATEWAY:
            replicas = 1
        sharding_enabled = shards and shards > 1

        cuda_device_map = None
        if self.args.env or os.environ.get('CUDA_VISIBLE_DEVICES', '').startswith('RR'):
            args_env = self.args.env or {}
            cuda_visible_devices = args_env.get(
                'CUDA_VISIBLE_DEVICES'
            ) or os.environ.get('CUDA_VISIBLE_DEVICES', None)

            cuda_device_map = Deployment._roundrobin_cuda_device(
                cuda_visible_devices, replicas
            )

        all_shard_pod_ports = self.args.peer_ports or {'0': []}
        if isinstance(all_shard_pod_ports, str):
            all_shard_pod_ports = json.loads(all_shard_pod_ports)
        if isinstance(all_shard_pod_ports, list):
            # it is a single shard
            all_shard_pod_ports = {'0': all_shard_pod_ports}

        peer_ports_all_shards = {}
        for k, v in all_shard_pod_ports.items():
            peer_ports_all_shards[str(k)] = v

        if self.args.stateful and len(peer_ports_all_shards.keys()) < shards:
            raise ValueError(
                'The configuration of `peer_ports` does not match the number of shards requested'
            )

        for shard_id in range(shards):
            peer_ports = peer_ports_all_shards.get(str(shard_id), [])
            if len(peer_ports) > 0 and len(peer_ports) != replicas:
                raise ValueError(
                    'peer-ports argument does not match number of replicas, it will be ignored'
                )
            elif len(peer_ports) == 0:
                peer_ports = [random_port() for _ in range(replicas)]

            replica_args = []

            for replica_id, peer_port in zip(range(replicas), peer_ports):
                _args = copy.deepcopy(self.args)
                if self.args.deployment_role == DeploymentRoleType.GATEWAY:
                    _args.replicas = replicas
                _args.shard_id = shard_id
                _args.replica_id = replica_id
                # for gateway pods, the pod role shouldn't be changed
                if _args.pod_role != PodRoleType.GATEWAY:
                    _args.pod_role = PodRoleType.WORKER
                    if len(self.args.host) == replicas:
                        _args.host = self.args.host[replica_id]
                    else:
                        _args.host = self.args.host[0]
                else:
                    _args.host = self.args.host

                if cuda_device_map:
                    _args.env = _args.env or {}
                    _args.env['CUDA_VISIBLE_DEVICES'] = str(cuda_device_map[replica_id])

                if _args.name:
                    _args.name += (
                        f'/shard-{shard_id}/rep-{replica_id}'
                        if sharding_enabled
                        else f'/rep-{replica_id}'
                    )
                else:
                    _args.name = f'{replica_id}'

                # the gateway needs to respect the assigned port
                if self.args.deployment_role == DeploymentRoleType.GATEWAY:
                    _args.port = self.args.port

                elif not self.external:
                    if shards == 1 and replicas == 1:
                        if len(_args.protocol) > 1 and self._include_gateway:
                            _args.port = [
                                random_port() for _ in range(len(self.args.protocol))
                            ]
                        else:
                            _args.port = self.args.port
                        _args.port_monitoring = self.args.port_monitoring

                    elif shards == 1:
                        _args.port_monitoring = (
                            random_port()
                            if replica_id >= len(self.args.all_port_monitoring)
                            else self.args.all_port_monitoring[replica_id]
                        )
                        # if there are no shards/replicas, we dont need to distribute ports randomly
                        # we should rather use the pre assigned one
                        _args.port = [
                            random_port() for _ in range(len(self.args.protocol))
                        ]
                        _args.port[0] = peer_port
                    elif shards > 1:
                        port_monitoring_index = (
                            replica_id + replicas * shard_id + 1
                        )  # the first index is for the head
                        _args.port_monitoring = (
                            random_port()
                            if port_monitoring_index
                            >= len(self.args.all_port_monitoring)
                            else self.args.all_port_monitoring[
                                port_monitoring_index
                            ]  # we skip the head port here
                        )
                        _args.port = [peer_port]
                    else:
                        _args.port = [peer_port]
                        _args.port_monitoring = random_port()

                else:
                    _args.port = [self.ext_repl_ports[replica_id]]
                    _args.host = self.ext_repl_hosts[replica_id]
                    _args.scheme = self.ext_repl_schemes[replica_id]
                    _args.tls = self.ext_repl_tls[replica_id]

                # pod workspace if not set then derive from workspace
                if not _args.workspace:
                    _args.workspace = self.args.workspace
                replica_args.append(_args)
            result[shard_id] = replica_args

        return result

    @staticmethod
    def _set_uses_before_after_args(args: Namespace, entity_type: str) -> Namespace:
        _args = copy.deepcopy(args)
        _args.pod_role = PodRoleType.WORKER
        _args.host = _args.host[0] or __default_host__
        _args.port = [random_port()]

        if _args.name:
            _args.name += f'/{entity_type}-0'
        else:
            _args.name = f'{entity_type}-0'

        if 'uses_before' == entity_type:
            _args.uses_requests = None
            _args.uses = args.uses_before or __default_executor__
        elif 'uses_after' == entity_type:
            _args.uses_requests = None
            _args.uses = args.uses_after or __default_executor__
        else:
            raise ValueError(
                f'uses_before/uses_after pod does not support type {entity_type}'
            )

        # pod workspace if not set then derive from workspace
        if not _args.workspace:
            _args.workspace = args.workspace
        return _args

    def _parse_base_deployment_args(self, args):
        parsed_args = {
            'head': None,
            'uses_before': None,
            'uses_after': None,
            'gateway': None,
            'pods': {},
        }

        if self.args.stateful and self.args.replicas in [1, 2]:
            self.logger.debug(
                'Stateful Executor is not recommended to be used less than 3 replicas'
            )

        if self.args.stateful and self.args.workspace is None:
            raise ValueError(
                'Stateful Executors need to be provided `workspace` when used in a Deployment'
            )

        # a gateway has no heads and uses
        # also there a no heads created, if there are no shards
        if self.role != DeploymentRoleType.GATEWAY and getattr(args, 'shards', 1) > 1:
            if (
                getattr(args, 'uses_before', None)
                and args.uses_before != __default_executor__
            ):
                uses_before_args = self._set_uses_before_after_args(
                    args, entity_type='uses_before'
                )
                parsed_args['uses_before'] = uses_before_args
                args.uses_before_address = (
                    f'{uses_before_args.host}:{uses_before_args.port[0]}'
                )
            if (
                getattr(args, 'uses_after', None)
                and args.uses_after != __default_executor__
            ):
                uses_after_args = self._set_uses_before_after_args(
                    args, entity_type='uses_after'
                )
                parsed_args['uses_after'] = uses_after_args
                args.uses_after_address = (
                    f'{uses_after_args.host}:{uses_after_args.port[0]}'
                )

            parsed_args['head'] = Deployment._copy_to_head_args(args)

        parsed_args['pods'] = self._set_pod_args()

        if parsed_args['head'] is not None:
            connection_list = defaultdict(list)

            for shard_id in parsed_args['pods']:
                for pod_idx, pod_args in enumerate(parsed_args['pods'][shard_id]):
                    worker_host = self.get_worker_host(pod_args, self._is_docker, False)
                    connection_list[shard_id].append(
                        f'{worker_host}:{pod_args.port[0]}'
                    )
            parsed_args['head'].connection_list = json.dumps(connection_list)

        return parsed_args

    @property
    def _mermaid_str(self) -> List[str]:
        """String that will be used to represent the Deployment graphically when `Flow.plot()` is invoked.
        It does not include used_before/uses_after


        .. # noqa: DAR201
        """
        mermaid_graph = []
        secret = '&ltsecret&gt'
        if self.role != DeploymentRoleType.GATEWAY and not self.external:
            mermaid_graph = [f'subgraph {self.name};', '\ndirection LR;\n']

            uses_before_name = (
                self.uses_before_args.name
                if self.uses_before_args is not None
                else None
            )
            uses_before_uses = (
                replace_secret_of_hub_uri(self.uses_before_args.uses, secret)
                if self.uses_before_args is not None
                else None
            )
            uses_after_name = (
                self.uses_after_args.name if self.uses_after_args is not None else None
            )
            uses_after_uses = (
                replace_secret_of_hub_uri(self.uses_after_args.uses, secret)
                if self.uses_after_args is not None
                else None
            )
            shard_names = []
            if len(self.pod_args['pods']) > 1:
                # multiple shards
                for shard_id, pod_args in self.pod_args['pods'].items():
                    shard_name = f'{self.name}/shard-{shard_id}'
                    shard_names.append(shard_name)
                    shard_mermaid_graph = [
                        f'subgraph {shard_name};',
                        '\ndirection TB;\n',
                    ]
                    names = [
                        args.name for args in pod_args
                    ]  # all the names of each of the replicas
                    uses = [
                        args.uses for args in pod_args
                    ]  # all the uses should be the same but let's keep it this
                    # way
                    for rep_i, (name, use) in enumerate(zip(names, uses)):
                        escaped_uses = f'"{replace_secret_of_hub_uri(use, secret)}"'
                        shard_mermaid_graph.append(f'{name}[{escaped_uses}]:::pod;')
                    shard_mermaid_graph.append('end;')
                    shard_mermaid_graph = [
                        node.replace(';', '\n') for node in shard_mermaid_graph
                    ]
                    mermaid_graph.extend(shard_mermaid_graph)
                    mermaid_graph.append('\n')
                if uses_before_name is not None:
                    for shard_name in shard_names:
                        escaped_uses_before_uses = (
                            f'"{replace_secret_of_hub_uri(uses_before_uses, secret)}"'
                        )
                        mermaid_graph.append(
                            f'{self.args.name}-head[{escaped_uses_before_uses}]:::HEADTAIL --> {shard_name};'
                        )
                if uses_after_name is not None:
                    for shard_name in shard_names:
                        escaped_uses_after_uses = f'"{uses_after_uses}"'
                        mermaid_graph.append(
                            f'{shard_name} --> {self.args.name}-tail[{escaped_uses_after_uses}]:::HEADTAIL;'
                        )
            else:
                # single shard case, no uses_before or uses_after_considered
                pod_args = list(self.pod_args['pods'].values())[0][0]
                uses = f'"{replace_secret_of_hub_uri(pod_args.uses, secret)}"'

                # just put the replicas in parallel
                if pod_args.replicas > 1:
                    for rep_i in range(pod_args.replicas):
                        mermaid_graph.append(
                            f'{pod_args.name}/rep-{rep_i}["{uses}"]:::pod;'
                        )
                else:
                    mermaid_graph.append(f'{pod_args.name}["{uses}"]:::pod;')

            mermaid_graph.append('end;')
        return mermaid_graph

    def block(
        self,
        stop_event: Optional[Union['threading.Event', 'multiprocessing.Event']] = None,
    ):
        """Block the Deployment until `stop_event` is set or user hits KeyboardInterrupt

        :param stop_event: a threading event or a multiprocessing event that once set will resume the control flow
            to main thread.
        """

        def _reload_deployment(changed_file):
            self.logger.info(
                f'change in Executor configuration YAML {changed_file} observed, reloading Executor deployment'
            )
            self.__exit__(None, None, None)
            new_deployment = Deployment(
                self.args, self.needs, include_gateway=self._include_gateway
            )
            self.__dict__ = new_deployment.__dict__
            self.__enter__()

        try:
            watch_changes = self.args.reload

            if watch_changes and self._is_executor_from_yaml:
                with ImportExtensions(
                    required=True,
                    help_text='''reload requires watchfiles dependency to be installed. You can run `pip install 
                    watchfiles''',
                ):
                    from watchfiles import watch

                new_stop_event = stop_event or threading.Event()
                if self._is_executor_from_yaml:
                    for changes in watch(*[self.args.uses], stop_event=new_stop_event):
                        for _, changed_file in changes:
                            _reload_deployment(self.args.uses)
            else:
                wait_event = stop_event
                if not wait_event:
                    self._stop_event = threading.Event()
                    wait_event = self._stop_event
                if not __windows__:
                    wait_event.wait()
                else:
                    while True:
                        if wait_event.is_set():
                            break
                        time.sleep(0.5)
        except KeyboardInterrupt:
            pass

    def _get_summary_table(self, all_panels: List[Panel]):
        address_table = self._init_table()

        if not isinstance(self.protocol, list):
            _protocols = [str(self.protocol)]
        else:
            _protocols = [str(_p) for _p in self.protocol]

        if not isinstance(self.first_pod_args.port, list):
            _ports = [self.first_pod_args.port]
        else:
            _ports = [str(_p) for _p in self.first_pod_args.port]

        swagger_ui_link = None
        redoc_link = None
        for _port, _protocol in zip(_ports, _protocols):
            address_table.add_row(':chains:', 'Protocol', _protocol)

            _protocol = _protocol.lower()
            address_table.add_row(
                ':house:',
                'Local',
                f'[link={_protocol}://{self.host}:{_port}]{self.host}:{_port}[/]',
            )
            address_table.add_row(
                ':lock:',
                'Private',
                f'[link={_protocol}://{self.address_private}:{_port}]{self.address_private}:{_port}[/]',
            )

            if self.address_public:
                address_table.add_row(
                    ':earth_africa:',
                    'Public',
                    f'[link={_protocol}://{self.address_public}:{_port}]{self.address_public}:{_port}[/]',
                )

            if _protocol == ProtocolType.HTTP.to_string().lower():
                swagger_ui_link = f'[link={_protocol}://{self.host}:{_port}/docs]{self.host}:{_port}/docs'
                redoc_link = f'[link={_protocol}://{self.host}:{_port}/redoc]{self.host}:{_port}/redoc'

        all_panels.append(
            Panel(
                address_table,
                title=':link: [b]Endpoint[/]',
                expand=False,
            )
        )

        if ProtocolType.HTTP.to_string().lower() in [p.lower() for p in _protocols]:
            http_ext_table = self._init_table()
            http_ext_table.add_row(':speech_balloon:', 'Swagger UI', swagger_ui_link)

            http_ext_table.add_row(':books:', 'Redoc', redoc_link)

            all_panels.append(
                Panel(
                    http_ext_table,
                    title=':gem: [b]HTTP extension[/]',
                    expand=False,
                )
            )

        if self.args.monitoring:
            monitor_ext_table = self._init_table()

            for replica in self.pod_args['pods'][0]:
                monitor_ext_table.add_row(
                    ':flashlight:',  # upstream issue: they dont have :torch: emoji, so we use :flashlight:
                    # to represent observability of Prometheus (even they have :torch: it will be a war
                    # between AI community and Cloud-native community fighting on this emoji)
                    replica.name,
                    f'...[b]:{replica.port_monitoring}[/]',
                )

                all_panels.append(
                    Panel(
                        monitor_ext_table,
                        title=':gem: [b]Prometheus extension[/]',
                        expand=False,
                    )
                )

        return all_panels

    @property
    def _docker_compose_address(self):
        from jina.orchestrate.deployments.config.docker_compose import port
        from jina.orchestrate.deployments.config.helper import to_compatible_name

        if self.external:
            docker_compose_address = [f'{self.protocol}://{self.host}:{self.port}']
        elif self.head_args:
            docker_compose_address = [
                f'{to_compatible_name(self.head_args.name)}:{port}'
            ]
        else:
            if self.args.replicas == 1 or self.name == 'gateway':
                docker_compose_address = [f'{to_compatible_name(self.name)}:{port}']
            else:
                docker_compose_address = []
                for rep_id in range(self.args.replicas):
                    node_name = f'{self.name}/rep-{rep_id}'
                    docker_compose_address.append(
                        f'{to_compatible_name(node_name)}:{port}'
                    )
        return docker_compose_address

    def _to_docker_compose_config(self, deployments_addresses=None):
        from jina.orchestrate.deployments.config.docker_compose import (
            DockerComposeConfig,
        )

        docker_compose_deployment = DockerComposeConfig(
            args=self.args, deployments_addresses=deployments_addresses
        )
        return docker_compose_deployment.to_docker_compose_config()

    def _inner_gateway_to_docker_compose_config(self):
        from jina.orchestrate.deployments.config.docker_compose import (
            DockerComposeConfig,
        )

        self.pod_args['gateway'].port = self.pod_args['gateway'].port or [random_port()]
        cargs = copy.deepcopy(self.pod_args['gateway'])
        cargs.uses = __default_grpc_gateway__
        cargs.graph_description = (
            f'{{"{self.name}": ["end-gateway"], "start-gateway": ["{self.name}"]}}'
        )

        docker_compose_deployment = DockerComposeConfig(
            args=cargs,
            deployments_addresses={self.name: self._docker_compose_address},
        )
        return docker_compose_deployment.to_docker_compose_config()

    def to_docker_compose_yaml(
        self,
        output_path: Optional[str] = None,
        network_name: Optional[str] = None,
    ):
        """
        Converts a Jina Deployment into a Docker compose YAML file

        If you don't want to rebuild image on Executor Hub,
        you can set `JINA_HUB_NO_IMAGE_REBUILD` environment variable.

        :param output_path: The path to dump the YAML file to
        :param network_name: The name of the network that will be used by the deployment
        """
        import yaml

        output_path = output_path or 'docker-compose.yml'
        network_name = network_name or 'jina-network'

        docker_compose_dict = {
            'version': '3.3',
            'networks': {network_name: {'driver': 'bridge'}},
        }
        services = {}

        service_configs = self._to_docker_compose_config()

        for service_name, service in service_configs:
            service['networks'] = [network_name]
            services[service_name] = service

        if self._include_gateway:
            service_configs = self._inner_gateway_to_docker_compose_config()

            for service_name, service in service_configs:
                service['networks'] = [network_name]
                services[service_name] = service

        docker_compose_dict['services'] = services
        with open(output_path, 'w+', encoding='utf-8') as fp:
            yaml.dump(docker_compose_dict, fp, sort_keys=False)

        command = (
            'docker-compose up'
            if output_path is None
            else f'docker-compose -f {output_path} up'
        )

        self.logger.info(
            f'Docker Compose file has been created under [b]{output_path}[/b]. You can use it by running [b]{command}[/b]'
        )

    def _to_kubernetes_yaml(
        self,
        output_base_path: str,
        k8s_namespace: Optional[str] = None,
        k8s_deployments_addresses: Optional[Dict] = None,
    ):
        import yaml

        from jina.orchestrate.deployments.config.k8s import K8sDeploymentConfig

        if self.external:
            self.logger.warning(
                'The Deployment is external, cannot create YAML deployment files'
            )
            return

        if self.args.name == 'gateway':
            if self.args.default_port:
                from jina.serve.networking import GrpcConnectionPool

                self.args.port = [
                    GrpcConnectionPool.K8S_PORT + i
                    for i in range(len(self.args.protocol))
                ]
                self.first_pod_args.port = [
                    GrpcConnectionPool.K8S_PORT + i
                    for i in range(len(self.args.protocol))
                ]

                self.args.port_monitoring = GrpcConnectionPool.K8S_PORT_MONITORING
                self.first_pod_args.port_monitoring = (
                    GrpcConnectionPool.K8S_PORT_MONITORING
                )

                self.args.default_port = False

            self.args.deployments_addresses = k8s_deployments_addresses
        else:
            if len(self.args.protocol) > 1 and len(self.args.port) != len(
                self.args.protocol
            ):
                from jina.serve.networking import GrpcConnectionPool

                self.args.port = [
                    GrpcConnectionPool.K8S_PORT + i
                    for i in range(len(self.args.protocol))
                ]
        k8s_deployment = K8sDeploymentConfig(
            args=self.args, k8s_namespace=k8s_namespace
        )

        configs = k8s_deployment.to_kubernetes_yaml()

        for name, k8s_objects in configs:
            filename = os.path.join(output_base_path, f'{name}.yml')
            os.makedirs(output_base_path, exist_ok=True)
            with open(filename, 'w+', encoding='utf-8') as fp:
                for i, k8s_object in enumerate(k8s_objects):
                    yaml.dump(k8s_object, fp)
                    if i < len(k8s_objects) - 1:
                        fp.write('---\n')

    def to_kubernetes_yaml(
        self,
        output_base_path: str,
        k8s_namespace: Optional[str] = None,
    ):
        """
        Convert a Jina Deployment into a set of YAML deployments to deploy in Kubernetes.

        If you don't want to rebuild image on Executor Hub,
        you can set `JINA_HUB_NO_IMAGE_REBUILD` environment variable.

        :param output_base_path: The base path where to dump all the YAML files
        :param k8s_namespace: The name of the k8s namespace to set for the configurations. If None, the name of the Flow will be used.
        """
        k8s_namespace = k8s_namespace or 'default'
        # the Deployment conversion needs to be done in a version without Gateway included. Deployment does quite some changes to its args
        # to let some of the args (like ports) go to the gateway locally. In Kubernetes, we want no gateway
        self._to_kubernetes_yaml(
            output_base_path=output_base_path,
            k8s_namespace=k8s_namespace,
        )
        self.logger.info(
            f'K8s YAML files have been created under [b]{output_base_path}[/]. You can use it by running [b]kubectl apply -R -f {output_base_path}[/]'
        )

    to_k8s_yaml = to_kubernetes_yaml



from jina.serve.runtimes.servers import BaseServer
from aiohttp import web


class LoadBalancingServer(BaseServer):
    """Base FastAPI server. Implement this abstract class in-case you want to build a fastapi-based server by
    implementing the `app` property. This property should return a fastapi app. The base Gateway will handle starting
    a server and serving the application using that server."""

    def __init__(
            self,
            **kwargs
    ):
        """Initialize the LoadBalancingServer
        :param kwargs: keyword args
        """
        super().__init__(**kwargs)
        # get server list from args
        self._server_exit = False

    async def handle_request(self, request):
        """Method called to handle requests coming to the LoadBalancer
        :param request: request to handle
        :return: the response to the request
        """
        return await self._request_handler._load_balance(request)

    async def setup_server(self):
        """
        Initialize and return server
        """
        self.logger.debug(f'Setting up LoadBalancer server')
        self.app = web.Application()
        self.app.router.add_route('*', '/{path:.*}', self.handle_request)
        self.logger.debug(f'LoadBalancer server setup successful')

    async def run_server(self):
        """Run HTTP server forever"""
        await web._run_app(
            app=self.app,
            host=self.host,
            port=self.port,
        )

    async def shutdown(self):
        """Shutdown the server and free other allocated resources, e.g, streamer object, health check service, ..."""
        self.logger.debug(f'Shutting down server')
        self._server_exit = True
        await super().shutdown()
        await self._request_handler.close()
        self.logger.debug(f'Server shutdown finished')

    @property
    def _should_exit(self):
        """Property describing if server is ready to exit
        :return: boolean indicating if Server ready to exit
        """
        return self._server_exit

    @property
    def should_exit(self):
        """Property describing if server is ready to exit
        :return: boolean indicating if Server ready to exit
        """
        return self._should_exit



# Copyright © 2020, [sysid](https://sysid.github.io/).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import enum
import io
import json
import logging
import re
from datetime import datetime
from functools import partial
from typing import Any, AsyncIterable, Callable, Coroutine, Dict, Optional, Union

import anyio
from starlette.background import BackgroundTask
from starlette.concurrency import iterate_in_threadpool
from starlette.responses import Response
from starlette.types import Receive, Scope, Send

_log = logging.getLogger(__name__)


# https://stackoverflow.com/questions/58133694/graceful-shutdown-of-uvicorn-starlette-app-with-websockets
class AppStatus:
    """helper for monkey-patching the signal-handler of uvicorn"""

    should_exit = False
    should_exit_event: Union[anyio.Event, None] = None

    @staticmethod
    def handle_exit(*args, **kwargs):
        """
        signal handler for uvicorn server
        :param args: args
        :param kwargs: kwargs
        """
        # set bool flag before checking the event to avoid race condition
        AppStatus.should_exit = True
        # Check if event has been initialized, if so notify listeners
        if AppStatus.should_exit_event is not None:
            AppStatus.should_exit_event.set()
        original_handler(*args, **kwargs)


try:
    from uvicorn.main import Server

    original_handler = Server.handle_exit
    Server.handle_exit = AppStatus.handle_exit  # type: ignore

    def unpatch_uvicorn_signal_handler():
        """restores original signal-handler and rolls back monkey-patching.
        Normally this should not be necessary.
        """
        Server.handle_exit = original_handler

except ModuleNotFoundError:
    _log.debug("Uvicorn not used.")


class SseState(enum.Enum):
    """SSE connection state"""

    CONNECTING = 0
    OPENED = 1
    CLOSED = 2


class ServerSentEvent:
    """Class to manage Server-Sent Events"""

    def __init__(
        self,
        data: Optional[Any] = None,
        *,
        event: Optional[str] = None,
        id: Optional[int] = None,
        retry: Optional[int] = None,
        comment: Optional[str] = None,
        sep: Optional[str] = None,
    ) -> None:
        """Send data using EventSource protocol # noqa: DAR101

        :param data: The data field for the message.
        :param event: The event's type. If this is specified, an event will be dispatched on the browser to the listener for the specified event name; the web site would use addEventListener() to listen for named events. The default event type is "message".
        :param id: The event ID to set the EventSource object's last event ID value to.
        :param retry: The reconnection time to use when attempting to send the event. [What code handles this?] This must be an integer, specifying the reconnection time in milliseconds. If a non-integer value is specified, the field is ignored.
        :param comment: A colon as the first character of a line is essence a comment, and is ignored. Usually used as a ping message to keep connecting. If set, this will be a comment message.
        :param sep: The separator between lines. Defaults to "\r\n".
        """
        self.data = data
        self.event = event
        self.id = id
        self.retry = retry
        self.comment = comment
        self.DEFAULT_SEPARATOR = "\r\n"
        self.LINE_SEP_EXPR = re.compile(r"\r\n|\r|\n")
        self._sep = sep if sep is not None else self.DEFAULT_SEPARATOR

    def encode(self) -> bytes:
        """
        Encode the message into bytes

        :return: The encoded message
        """
        buffer = io.StringIO()
        if self.comment is not None:
            for chunk in self.LINE_SEP_EXPR.split(str(self.comment)):
                buffer.write(f": {chunk}")
                buffer.write(self._sep)

        if self.id is not None:
            buffer.write(self.LINE_SEP_EXPR.sub("", f"id: {self.id}"))
            buffer.write(self._sep)

        if self.event is not None:
            buffer.write(self.LINE_SEP_EXPR.sub("", f"event: {self.event}"))
            buffer.write(self._sep)

        if self.data is not None:
            for chunk in self.LINE_SEP_EXPR.split(json.dumps(self.data)):
                buffer.write(f"data: {chunk}")
                buffer.write(self._sep)

        if self.retry is not None:
            if not isinstance(self.retry, int):
                raise TypeError("retry argument must be int")
            buffer.write(f"retry: {self.retry}")
            buffer.write(self._sep)

        buffer.write(self._sep)
        return buffer.getvalue().encode("utf-8")


def ensure_bytes(data: Union[bytes, dict, ServerSentEvent, Any]) -> bytes:
    """
    helper to convert data to bytes
    :param data: data to convert
    :return: bytes
    """
    if isinstance(data, bytes):
        return data
    elif isinstance(data, ServerSentEvent):
        return data.encode()
    elif isinstance(data, dict):
        return ServerSentEvent(**data).encode()
    else:
        return ServerSentEvent(str(data)).encode()


class EventSourceResponse(Response):
    """Implements the ServerSentEvent Protocol:
    https://www.w3.org/TR/2009/WD-eventsource-20090421/

    Responses must not be compressed by middleware in order to work.
    implementation based on Starlette StreamingResponse
    """

    DEFAULT_PING_INTERVAL = 15

    # noinspection PyMissingConstructor
    def __init__(
        self,
        content: Any,
        status_code: int = 200,
        headers: Optional[Dict] = None,
        media_type: str = "text/event-stream",
        background: Optional[BackgroundTask] = None,
        ping: Optional[int] = None,
        sep: Optional[str] = None,
        ping_message_factory: Optional[Callable[[], ServerSentEvent]] = None,
        data_sender_callable: Optional[
            Callable[[], Coroutine[None, None, None]]
        ] = None,
    ) -> None:
        self.sep = sep
        self.ping_message_factory = ping_message_factory
        if isinstance(content, AsyncIterable):
            self.body_iterator = (
                content
            )  # type: AsyncIterable[Union[Any,dict,ServerSentEvent]]
        else:
            self.body_iterator = iterate_in_threadpool(content)  # type: ignore
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type
        self.background = background  # type: ignore  # follows https://github.com/encode/starlette/blob/master/starlette/responses.py
        self.data_sender_callable = data_sender_callable

        _headers = {}
        if headers is not None:  # pragma: no cover
            _headers.update(headers)

        # mandatory for servers-sent events headers
        # allow cache control header to be set by user to support fan out proxies
        # https://www.fastly.com/blog/server-sent-events-fastly
        _headers.setdefault("Cache-Control", "no-cache")
        _headers["Connection"] = "keep-alive"
        _headers["X-Accel-Buffering"] = "no"

        self.init_headers(_headers)

        self.ping_interval = self.DEFAULT_PING_INTERVAL if ping is None else ping
        self.active = True

        self._ping_task = None
        self._send_lock = anyio.Lock()

    @staticmethod
    async def listen_for_disconnect(receive: Receive) -> None:
        """
        Listen for the client disconnecting
        :param receive: receive channel
        """
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                _log.debug("Got event: http.disconnect. Stop streaming.")
                break

    @staticmethod
    async def listen_for_exit_signal() -> None:
        """
        Listen for the exit signal
        """
        # Check if should_exit was set before anybody started waiting
        if AppStatus.should_exit:
            return

        # Setup an Event
        if AppStatus.should_exit_event is None:
            AppStatus.should_exit_event = anyio.Event()

        # Check if should_exit got set while we set up the event
        if AppStatus.should_exit:
            return

        # Await the event
        await AppStatus.should_exit_event.wait()

    async def stream_response(self, send) -> None:
        """
        Stream the response
        :param send: send channel
        """
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for data in self.body_iterator:
            chunk = ensure_bytes(data)
            _log.debug(f"chunk: {chunk.decode()}")
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Stream the response
        :param scope: scope
        :param receive: receive channel
        :param send: send channel
        """

        async def safe_send(message):
            async with self._send_lock:
                return await send(message)

        async with anyio.create_task_group() as task_group:
            # https://trio.readthedocs.io/en/latest/reference-core.html#custom-supervisors
            async def wrap(func: Callable[[], Coroutine[None, None, None]]) -> None:
                await func()
                # noinspection PyAsyncCall
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.stream_response, safe_send))
            task_group.start_soon(wrap, partial(self._ping, safe_send))
            task_group.start_soon(wrap, self.listen_for_exit_signal)

            if self.data_sender_callable:
                task_group.start_soon(self.data_sender_callable)

            await wrap(partial(self.listen_for_disconnect, receive))

        if self.background is not None:  # pragma: no cover, tested in StreamResponse
            await self.background()

    def enable_compression(self, force: bool = False) -> None:
        """
        Enable compression
        :param force: force compression
        """
        raise NotImplementedError

    @property
    def ping_interval(self) -> Union[int, float]:
        """
        Time interval between two ping massages

        :return: ping interval
        """
        return self._ping_interval

    @ping_interval.setter
    def ping_interval(self, value: Union[int, float]) -> None:
        """Setter for ping_interval property.

        :param value: interval in sec between two ping values.
        """

        if not isinstance(value, (int, float)):
            raise TypeError("ping interval must be int")
        if value < 0:
            raise ValueError("ping interval must be greater then 0")

        self._ping_interval = value

    async def _ping(self, send: Send) -> None:
        # Legacy proxy servers are known to, in certain cases, drop HTTP connections after a short timeout.
        # To protect against such proxy servers, authors can send a custom (ping) event
        # every 15 seconds or so.
        # Alternatively one can send periodically a comment line
        # (one starting with a ':' character)
        while self.active:
            await anyio.sleep(self._ping_interval)
            if self.ping_message_factory:
                assert isinstance(self.ping_message_factory, Callable)  # type: ignore  # https://github.com/python/mypy/issues/6864
            ping = (
                ServerSentEvent(comment=f"ping - {datetime.utcnow()}").encode()
                if self.ping_message_factory is None
                else ensure_bytes(self.ping_message_factory())
            )
            _log.debug(f"ping: {ping.decode()}")
            await send({"type": "http.response.body", "body": ping, "more_body": True})
